# %%

from os import listdir
from pathlib import Path
import joblib
import pandas as pd

import matplotlib.pyplot as plt
import seaborn as sns
import scipy.signal as dsp
import numpy as np


sns.set_style('ticks')
sns.set_context('poster')


# %%
INDIR = '/mnt/obob/staff/fschmidt/resting_tinnitus/data/data_meg/'
subject_ids = listdir(INDIR)


# for subject_id in subject_ids:

subject_id = subject_ids[2]

cur_data = joblib.load(list(Path(INDIR).glob(f'{subject_id}/{subject_id}__fft_method_irasa.dat'))[0])
freq_range_ap = cur_data['freq'] < 100
freq_range_p = np.logical_and(cur_data['freq'] < 40, cur_data['freq'] > 1)
# %%
plt.loglog(cur_data['freq'][freq_range_ap], cur_data['src']['label_tc_ap'].T[freq_range_ap, :])
# %%
plt.plot(cur_data['freq'][freq_range_p], cur_data['src']['label_tc_p'].T[freq_range_p, :])
# %%


# %% find peaks irasa style
def get_peak_params(
    periodic_spectrum,
    freqs,
    ch_names=[],
    smoothing_window=2,
    cut_spectrum=(1, 40),
    peak_threshold=1,
    peak_width_limits=(1, 12),
):
    """
    This function can be used to extract peak parameters from the periodic spectrum extracted from IRASA.
    The algorithm works by smoothing the spectrum, zeroing out negative values and
    extracting peaks based on user specified parameters.

    Parameters: periodic_spectrum : 1d or 2d array
                    Power values for the periodic spectrum extracted using IRASA shape(channel x frequency)
                freqs : 1d array
                    Frequency values for the periodic spectrum
                ch_names: list, optional, default: []
                    Channel names ordered according to the periodic spectrum.
                    If empty channel names are given as numbers in ascending order.
                smoothing window : int, optional, default: 2
                    Smoothing window in Hz handed over to the savitzky-golay filter.
                cut_spectrum : tuple of (float, float), optional, default (1, 40)
                    Cut the periodic spectrum to limit peak finding to a sensible range
                peak_threshold: float, optional, default: 2.0
                    Relative threshold for detecting peaks. This threshold is defined in
                    relative units of the periodic spectrum
                peak_width_limits : tuple of (float, float), optional, default (1, 12)
                    Limits on possible peak width, in Hz, as (lower_bound, upper_bound)

    Returns:    df_peaks: DataFrame
                    DataFrame containing the center frequency, bandwidth and peak height for each channel


    """

    polyorder = 1  # polyorder for smoothing

    if np.isnan(periodic_spectrum).sum() > 0:
        raise ValueError('peak width detection does not work properly with nans')

    freq_step = freqs[1] - freqs[0]
    window_length = int(smoothing_window // freq_step)

    # generate channel names if not given
    if len(ch_names) == 0:
        ch_names = np.arange(periodic_spectrum.shape[0])

    # cut data
    if cut_spectrum != None:
        freq_range = np.logical_and(freqs > cut_spectrum[0], freqs < cut_spectrum[1])
        freqs = freqs[freq_range]
        periodic_spectrum = periodic_spectrum[:, freq_range]

    # filter signal to get a smoother spectrum for peak extraction
    filtered_spectrum = dsp.savgol_filter(periodic_spectrum, window_length=window_length, polyorder=polyorder)
    # zero out negative values
    filtered_spectrum[filtered_spectrum < 0] = 0

    # do peak finding on a channel by channel basis
    peak_list = []
    for ix, ch_name in enumerate(ch_names):
        peaks, peak_dict = dsp.find_peaks(
            filtered_spectrum[ix],
            width=peak_width_limits / freq_step,  # in frequency in hz
            prominence=peak_threshold * np.std(filtered_spectrum[ix]),
            rel_height=0.75,
        )  # relative peak height based on width
        cur_df = pd.DataFrame(
            {
                'ch_name': 'chx',
                'cf': freqs[peaks],
                'bw': peak_dict['widths'] * freq_step,
                'pw': peak_dict['prominences'],
            }
        )

        peak_list.append(
            pd.DataFrame(
                {
                    'ch_name': ch_name,
                    'cf': freqs[peaks],
                    'bw': peak_dict['widths'] * freq_step,
                    'pw': peak_dict['prominences'],
                }
            )
        )
    # combine & return
    df_peaks = pd.concat(peak_list)

    return df_peaks


# %%
periodic_spectrum = cur_data['src']['label_tc_p']
ch_names = cur_data['src']['label_info']['names_order_mne']
freqs = cur_data['freq']
smoothing_window = 1.5
cut_spectrum = (1, 40)
peak_threshold = 1
peak_width_limits = (0.5, 12)

df_peaks = get_peak_params(
    periodic_spectrum=periodic_spectrum,
    ch_names=ch_names,
    freqs=freqs,
    smoothing_window=smoothing_window,
    cut_spectrum=cut_spectrum,
    peak_threshold=peak_threshold,
    peak_width_limits=peak_width_limits,
)
# %%
polyorder = 1  # polyorder for smoothing
freq_step = freqs[1] - freqs[0]
distance = 1 / freq_step
window_length = int(smoothing_window // freq_step)
# cut data

if cut_spectrum != None:
    freq_range = np.logical_and(freqs > cut_spectrum[0], freqs < cut_spectrum[1])
    freqs = freqs[freq_range]
    periodic_spectrum = periodic_spectrum[:, freq_range]


# filter signal to get a smoother spectrum for peak extraction
filtered_spectrum = dsp.savgol_filter(periodic_spectrum, window_length=window_length, polyorder=polyorder)
# zero out negative values
filtered_spectrum[filtered_spectrum < 0] = 0

# NOTE: peak width cant work with nans -> add a check
ix = 1
peaks, peak_dict = dsp.find_peaks(
    filtered_spectrum[ix],
    width=peak_width_limits / freq_step,  # in frequency in hz
    prominence=peak_threshold * np.std(filtered_spectrum[ix]),
    rel_height=0.75,
)  # relative peak height
# %%
cur_df = pd.DataFrame(
    {'ch_name': 'chx', 'cf': freqs[peaks], 'bw': peak_dict['widths'] * freq_step, 'pw': peak_dict['prominences']}
)

# %%
fig, axes = plt.subplots(ncols=3, figsize=(12, 4))

# orignal
axes[0].plot(freqs, periodic_spectrum[ix, :])
axes[0].set_title('Original Spectrum')
axes[0].set_xlabel('Frequency (Hz)')
axes[0].set_ylabel('Power (a.u.)')
# filtered
axes[1].plot(freqs, filtered_spectrum[ix, :])
axes[1].set_title('Filtered Spectrum')
axes[1].set_xlabel('Frequency (Hz)')
axes[1].set_ylabel('Power (a.u.)')
# %filtered + peaks
axes[2].plot(freqs, filtered_spectrum[ix, :])
axes[2].set_title('Filtered Spectrum \n+ Peaks')
axes[2].set_xlabel('Frequency (Hz)')
axes[2].set_ylabel('Power (a.u.)')

for row in cur_df.iterrows():
    logical = np.logical_and(freqs > row[1]['cf'] - row[1]['bw'] / 2, freqs < row[1]['cf'] + row[1]['bw'] / 2)
    axes[2].fill_between(freqs, row[1]['pw'], where=logical, facecolor='red', alpha=0.5)


plt.tight_layout()

# %%
fig, ax = plt.subplots(figsize=(7, 7))
# %filtered + peaks
ax.plot(freqs, filtered_spectrum[ix, :])
ax.set_title('Filtered Spectrum \n+ Peaks')
ax.set_xlabel('Frequency (Hz)')
ax.set_ylabel('Power (a.u.)')

for row in cur_df.iterrows():
    logical = np.logical_and(freqs > row[1]['cf'] - row[1]['bw'] / 2, freqs < row[1]['cf'] + row[1]['bw'] / 2)
    # logical = np.logical_and(freqs > row[1]['cf'] - 1, freqs < row[1]['cf'] + 1)
    ax.fill_between(freqs, row[1]['pw'], where=logical, facecolor='red', alpha=0.5)


plt.tight_layout()
# %%
peak_dict
# %% now lets get band specific information

freq_range = (8, 12)
ch_names = cur_data['src']['label_info']['names_order_mne']


def get_band_info(df_peaks, freq_range, ch_names):
    """ """

    df_range = df_peaks.query(f'cf > {freq_range[0]}').query(f'cf < {freq_range[1]}')

    # we dont always get a peak in a queried range lets give those channels a nan
    missing_channels = list(set(ch_names).difference(df_range['ch_name'].unique()))
    missing_df = pd.DataFrame(np.nan, index=np.arange(len(missing_channels)), columns=['ch_name', 'cf', 'bw', 'pw'])
    missing_df['ch_name'] = missing_channels

    df_new_range = pd.concat([df_range, missing_df]).reset_index().drop(columns='index')

    return df_new_range


# %%
get_band_info(df_peaks, (8, 12), ch_names)
# %%



# %%
import joblib
from pathlib import Path
import os
import pandas as pd
import numpy as np

import matplotlib.pyplot as plt
import seaborn as sns

sns.set_style('ticks')
sns.set_context('poster')


base_p = '/'.join(os.path.abspath(os.path.dirname(__file__)).split('/')[:-1])

all_files = list(Path(base_p + '/data/sim_all_f/fixed').glob('fixed__*.dat'))
# %%

ap_list, p_list = [], []

for f in all_files:
    cur_f = joblib.load(f)

    drop_list = ['ch_name', 'sim_bw', 'sim_height', 'sim_b_proba', 'sim_knee']
    cur_p = pd.concat([cur_f['specparam']['fixed']['periodic'], cur_f['pyrasa']['fixed']['periodic']]).dropna(axis=1)

    cur_ap = pd.concat([cur_f['specparam']['fixed']['aperiodic'], cur_f['pyrasa']['fixed']['aperiodic']]).dropna(axis=1)

    ap_list.append(cur_ap)
    p_list.append(cur_p)

# %%
df_ap_sim = pd.concat(ap_list)

# %%
df_ap_sim['sim_exp'] = df_ap_sim['sim_exp'].abs()
df_ap_sim['Exponent Δ'] = np.abs(df_ap_sim['Exponent'] - df_ap_sim['sim_exp'])
# %%
f, ax = plt.subplots(figsize=(8, 8), nrows=2)
sns.lineplot(df_ap_sim, x='sim_freq', y='Exponent Δ', hue='method', legend=False, ax=ax[0])
ax[0].set_xlabel('Simulated Oscillation (Hz)')

sns.lineplot(df_ap_sim, x='sim_exp', y='Exponent Δ', hue='method', legend=False, ax=ax[1])

ax[1].set_xlabel('Simulated Exponent')

sns.despine()
f.tight_layout()

# %%
df_p_sim = pd.concat(p_list)
df_p_sim['sim_exp'] = df_p_sim['sim_exp'].abs()
# %%
df_nona = df_p_sim.copy().query('cf < 40')
df_nona['Center Frequency Δ'] = np.abs(df_nona['cf'] - df_nona['sim_freq'])

# %%
f, ax = plt.subplots(figsize=(4, 6))
sns.violinplot(
    df_nona,
    x='method',
    y='Center Frequency Δ',
    hue='method',
    # split=True,
    legend=False,
    ax=ax,
)

ax.set_xlabel('')
sns.despine()

# %%
f, ax = plt.subplots(figsize=(8, 8), nrows=2)
sns.lineplot(df_nona, x='sim_freq', y='Center Frequency Δ', hue='method', legend=False, ax=ax[0])
ax[0].set_xlabel('Simulated Oscillation (Hz)')

sns.lineplot(df_nona, x='sim_exp', y='Center Frequency Δ', hue='method', legend=False, ax=ax[1])

ax[1].set_xlabel('Simulated Exponent')

sns.despine()
f.tight_layout()
# %%

df_hp_sp = df_nona.query('method == "specparam"').query('pw > 1').copy()
df_ir = df_p_sim.query('method == "pyrasa"')

df_p_cmb = pd.concat([df_hp_sp, df_ir])

df_p_cmb['Center Frequency Δ'] = np.abs(df_p_cmb['cf'] - df_p_cmb['sim_freq'])

f, ax = plt.subplots(figsize=(4, 6))
sns.violinplot(
    df_p_cmb,
    x='method',
    y='Center Frequency Δ',
    hue='method',
    # split=True,
    legend=False,
    ax=ax,
)

ax.set_xlabel('')
sns.despine()

# %%
f, ax = plt.subplots(figsize=(8, 8), nrows=2)
sns.lineplot(df_p_cmb, x='sim_freq', y='Center Frequency Δ', hue='method', legend=False, ax=ax[0])
ax[0].set_xlabel('Simulated Oscillation')

sns.lineplot(df_p_cmb, x='sim_exp', y='Center Frequency Δ', hue='method', legend=False, ax=ax[1])

sns.despine()
f.tight_layout()


# %%
f, axes = plt.subplots(ncols=3, figsize=(14, 6))

sns.violinplot(
    df_ap_sim,
    x='method',
    y='Exponent Δ',
    hue='method',
    # split=True,
    legend=False,
    ax=axes[0],
)

sns.violinplot(
    df_nona,
    x='method',
    y='Center Frequency Δ',
    hue='method',
    # split=True,
    legend=False,
    ax=axes[1],
)

sns.violinplot(
    df_p_cmb,
    x='method',
    y='Center Frequency Δ',
    hue='method',
    # split=True,
    legend=False,
    ax=axes[2],
)

titles = ['', 'all peaks', 'specparam \n peaks > 1 (arbitrary)']
for ix, ax in enumerate(axes):
    ax.set_xlabel('')
    ax.set_title(titles[ix])

f.tight_layout()

sns.despine()
# %%



# %%
from neurodsp.sim import sim_powerlaw, sim_peak_oscillation, sim_combined, sim_knee

import joblib

import scipy.signal as dsp
import numpy as np
import pandas as pd
import os
import scipy.stats as sts

import matplotlib.pyplot as plt
import seaborn as sns

import matplotlib as mpl

new_rc_params = {'text.usetex': False, 'svg.fonttype': 'none'}
mpl.rcParams.update(new_rc_params)

sns.set_style('ticks')
sns.set_context('poster')

pal = [  #'#cccccc',
    '#969696',
    '#636363',
    '#252525',
]


n_secs = 60
fs = 500
kwargs_psd = {'nperseg': fs * 4, 'noverlap': fs * 2}

from pathlib import Path

base_p = '/'.join(os.path.abspath(os.path.dirname(__file__)).split('/')[:-1]) + '/data/sim_all_simple11/'


def get_simulations(folder):
    all_files = list(Path(base_p + folder).glob('*.dat'))

    ap_list, p_list = [], []
    for f in all_files:
        cur_f = joblib.load(f)

        cur_p = pd.concat([cur_f['specparam']['periodic'], cur_f['pyrasa']['periodic']])  # .dropna(axis=1)

        cur_ap = pd.concat([cur_f['specparam']['aperiodic'], cur_f['pyrasa']['aperiodic']])  # .dropna(axis=1)

        ap_list.append(cur_ap)
        p_list.append(cur_p)

    df_ap = pd.concat(ap_list)
    df_p = pd.concat(p_list)

    return df_ap, df_p


# %% Setting #1 - varying slope
exp_list = [-1.0, -1.5, -2]

kwargs_psd = {'nperseg': fs * 4, 'noverlap': fs * 2}
sigs = []
for exp in exp_list:
    sig = sim_powerlaw(n_seconds=n_secs, fs=fs, exponent=exp)
    f, psd = dsp.welch(sig, fs=fs, **kwargs_psd)
    df = pd.DataFrame({'Frequency (Hz)': f, 'Power': psd})
    df['exponent'] = np.abs(exp)
    sigs.append(df)

df_1 = pd.concat(sigs)


df_ap, df_p = get_simulations('exponent_delta')
# df_p = df_p.query('cf < 40').reset_index()
f, ax = plt.subplots(figsize=(15, 5), ncols=3)

sns.lineplot(
    df_1.query('`Frequency (Hz)` < 100'),
    legend=False,
    x='Frequency (Hz)',
    y='Power',
    hue='exponent',
    palette=pal,
    ax=ax[0],
)
ax[0].set_xscale('log')
ax[0].set_yscale('log')

ax[1].scatter(
    df_ap.query('method == "pyrasa"')['Exponent'],
    df_ap.query('method == "specparam"')['Exponent'],
)
ax[1].set_xlabel('Exponent \n (PyRASA)')
ax[1].set_ylabel('Exponent \n (Specparam)')
ax[1].set_xlim(0, 3)
ax[1].set_ylim(0, 3)

r, p = sts.pearsonr(
    df_ap.query('method == "pyrasa"')['Exponent'],
    df_ap.query('method == "specparam"')['Exponent'],
)
r, p = np.round(r, 2), np.round(p, 3)
ax[1].annotate(f'r = {r}', xy=(1.5, 0.6))
ax[1].annotate(f'p = {p}', xy=(1.5, 0.2))

n_runs = len(df_p['seed'].unique()) * len(df_p['sim_exp'].unique())
df_p = df_p.query('cf < 40')
df_p2 = df_p[['cf', 'pw', 'bw', 'method']].dropna(axis=0)
osc_per_run = pd.DataFrame(df_p2.groupby('method').apply(len) / n_runs).reset_index()
osc_per_run.columns = ['method', 'Oscillations detected \n (per Model fit)']
sns.barplot(
    osc_per_run,
    x='method',
    y='Oscillations detected \n (per Model fit)',
    hue='method',
    hue_order=['specparam', 'pyrasa'],
    palette='deep',
    ax=ax[2],
)
ax[2].set_xlabel('')
sns.despine()
f.tight_layout()

f.savefig('../results/exponent_delta_sim.svg')

# %% Setting #2 - varying knee
knee_list = [10.0, 20.0, 40.0]

sigs = []
for knee in knee_list:
    exp = -2
    sig = sim_knee(n_seconds=n_secs, fs=fs, exponent1=0, exponent2=exp, knee=knee ** np.abs(exp))
    f, psd = dsp.welch(sig, fs=fs, **kwargs_psd)
    df = pd.DataFrame({'Frequency (Hz)': f, 'Power': psd})
    df['knee'] = knee
    sigs.append(df)

df_1 = pd.concat(sigs)


df_ap, df_p = get_simulations('knee_delta')

# %%
f, ax = plt.subplots(figsize=(15, 5), ncols=3)

sns.lineplot(
    df_1.query('`Frequency (Hz)` < 100'), legend=False, x='Frequency (Hz)', y='Power', hue='knee', palette=pal, ax=ax[0]
)
ax[0].set_xscale('log')
ax[0].set_yscale('log')

ax[1].scatter(
    df_ap.query('method == "pyrasa"')['Knee Frequency (Hz)'],
    df_ap.query('method == "specparam"')['Knee Frequency (Hz)'],
)
ax[1].set_xlabel('Knee Frequency \n (Hz; PyRASA)')
ax[1].set_ylabel('Knee Frequency \n (Hz; Specparam)')
ax[1].set_xlim(0, 40)
ax[1].set_ylim(0, 40)


r, p = sts.pearsonr(
    df_ap.query('method == "pyrasa"')['Knee Frequency (Hz)'],
    df_ap.query('method == "specparam"')['Knee Frequency (Hz)'],
)
r, p = np.round(r, 2), np.round(p, 3)
ax[1].annotate(f'r = {r}', xy=(20, 8))
ax[1].annotate(f'p = {p}', xy=(20, 2))

f.tight_layout()
sns.despine()


n_runs = len(df_p['seed'].unique()) * len(df_p['sim_knee_freq'].unique())
df_p = df_p.query('cf < 40')
df_p2 = df_p[['cf', 'pw', 'bw', 'method']].dropna(axis=0)
osc_per_run = pd.DataFrame(df_p2.groupby('method').apply(len) / n_runs).reset_index()
osc_per_run.columns = ['method', 'Oscillations detected \n (per Model fit)']
sns.barplot(
    osc_per_run,
    x='method',
    y='Oscillations detected \n (per Model fit)',
    hue='method',
    hue_order=['specparam', 'pyrasa'],
    palette='deep',
    ax=ax[2],
)
ax[2].set_xlabel('')
sns.despine()
f.tight_layout()

f.savefig('../results/knee_delta_sim.svg')

# %% Setting #3 - varying exponent post knee
exp_list = [-1.0, -2.0, -3.0]

sigs = []
for exp in exp_list:
    sig = sim_knee(n_seconds=n_secs, fs=fs, exponent1=0, exponent2=exp, knee=15 ** np.abs(exp))
    f, psd = dsp.welch(sig, fs=fs, **kwargs_psd)
    df = pd.DataFrame({'Frequency (Hz)': f, 'Power': psd})
    df['exp'] = exp
    sigs.append(df)

df_1 = pd.concat(sigs)


# %%
f, ax = plt.subplots(figsize=(15, 5), ncols=3)

sns.lineplot(
    df_1.query('`Frequency (Hz)` < 100'), legend=False, x='Frequency (Hz)', y='Power', hue='exp', palette=pal, ax=ax[0]
)
ax[0].set_xscale('log')
ax[0].set_yscale('log')

df_ap, df_p = get_simulations('exponent_knee_delta')

ax[1].scatter(
    df_ap.query('method == "pyrasa"')['Exponent_2'],
    df_ap.query('method == "specparam"')['Exponent_2'],
)
ax[1].set_xlabel('Exponent \n (PyRASA)')
ax[1].set_ylabel('Exponent \n (Specparam)')
ax[1].set_xlim(1, 4)
ax[1].set_ylim(1, 4)

f.tight_layout()

r, p = sts.pearsonr(
    df_ap.query('method == "pyrasa"')['Exponent_2'],
    df_ap.query('method == "specparam"')['Exponent_2'],
)

r, p = np.round(r, 2), np.round(p, 3)
ax[1].annotate(f'r = {r}', xy=(2.5, 1.5))
ax[1].annotate(f'p = {p}', xy=(2.5, 1.1))


n_runs = len(df_p['seed'].unique()) * len(df_p['sim_knee_freq'].unique())
df_p = df_p.query('cf < 40')
df_p2 = df_p[['cf', 'pw', 'bw', 'method']].dropna(axis=0)
osc_per_run = pd.DataFrame(df_p2.groupby('method').apply(len) / n_runs).reset_index()
osc_per_run.columns = ['method', 'Oscillations detected \n (per Model fit)']
sns.barplot(
    osc_per_run,
    x='method',
    y='Oscillations detected \n (per Model fit)',
    hue='method',
    hue_order=['specparam', 'pyrasa'],
    palette='deep',
    ax=ax[2],
)
ax[2].set_xlabel('')
sns.despine()
f.tight_layout()

f.savefig('../results/exponent_knee_delta_sim.svg')

sns.despine()
# %% Setting #4 - freq_delta
sigs = []
osc_list = [10, 15, 20]
for freq in osc_list:
    comps = {
        'sim_powerlaw': {
            'exponent': -1.5,
        },
        'sim_oscillation': {'freq': freq},
    }
    sig = sim_combined(n_seconds=n_secs, fs=fs, components=comps)
    f, psd = dsp.welch(sig, fs=fs, **kwargs_psd)
    df = pd.DataFrame({'Frequency (Hz)': f, 'Power': psd})
    df['osc'] = freq
    sigs.append(df)

df_1 = pd.concat(sigs)


f, ax = plt.subplots(figsize=(15, 5), ncols=3)

sns.lineplot(
    df_1.query('`Frequency (Hz)` < 50').query('`Frequency (Hz)` > 1'),
    legend=False,
    x='Frequency (Hz)',
    y='Power',
    hue='osc',
    palette=pal,
    ax=ax[0],
)

ax[0].set_xscale('log')
ax[0].set_yscale('log')

df_ap, df_p = get_simulations('freq_delta')

n_runs = len(df_p['seed'].unique()) * len(df_p['sim_freq'].unique())
df_p = df_p.query('cf < 40')
df_p['Center Frequency Δ'] = np.abs(df_p['cf'] - df_p['sim_freq'])

sns.barplot(
    df_p,  # legend=False,
    x='method',
    y='Center Frequency Δ',
    hue='method',
    palette='deep',
    hue_order=['specparam', 'pyrasa'],
    ax=ax[1],
)
ax[1].set_xlabel('')


df_p2 = df_p[['cf', 'pw', 'bw', 'method']].dropna(axis=0)
osc_per_run = pd.DataFrame(df_p2.groupby('method').apply(len) / n_runs).reset_index()
osc_per_run.columns = ['method', 'Oscillations detected \n (per Model fit)']
sns.barplot(
    osc_per_run,
    x='method',
    y='Oscillations detected \n (per Model fit)',
    hue='method',
    hue_order=['specparam', 'pyrasa'],
    palette='deep',
    ax=ax[2],
)
ax[2].set_xlabel('')
sns.despine()
f.tight_layout()

f.savefig('../results/freq_delta_sim.svg')

# %% Setting #5 - freq_width_delta
sigs = []
bw_list = [1, 2, 3]
for bw in bw_list:
    height = 1.5
    freq = 10
    exp = -1.5
    peak_params = {'freq': freq, 'bw': bw, 'height': height}

    sig_ap = sim_powerlaw(n_seconds=n_secs, fs=fs, exponent=exp)
    sig = sim_peak_oscillation(sig_ap, fs=fs, **peak_params)

    f, psd = dsp.welch(sig, fs=fs, **kwargs_psd)
    df = pd.DataFrame({'Frequency (Hz)': f, 'Power': psd})
    df['bw'] = bw
    sigs.append(df)

df_1 = pd.concat(sigs)

f, ax = plt.subplots(figsize=(15, 5), ncols=3)

sns.lineplot(
    df_1.query('`Frequency (Hz)` < 50').query('`Frequency (Hz)` > 2'),
    legend=False,
    x='Frequency (Hz)',
    y='Power',
    hue='bw',
    palette=pal,
    ax=ax[0],
)

ax[0].set_xscale('log')
ax[0].set_yscale('log')


df_ap, df_p = get_simulations('freq_width_delta')
df_p = df_p.reset_index()
n_runs = len(df_p['seed'].unique()) * len(df_p['sim_freq'].unique())
df_p = df_p.query('cf < 40')
df_p['Bandwidth Δ'] = np.abs(df_p['bw'] - df_p['sim_bw'])

sns.barplot(
    df_p,  # legend=False,
    x='method',
    y='Bandwidth Δ',
    hue='method',
    palette='deep',
    ax=ax[1],
)
ax[1].set_xlabel('')

df_p2 = df_p[['cf', 'pw', 'bw', 'method']].dropna(axis=0)
osc_per_run = pd.DataFrame(df_p2.groupby('method').apply(len) / n_runs).reset_index()
osc_per_run.columns = ['method', 'Oscillations detected \n (per Model fit)']
sns.barplot(
    osc_per_run,
    x='method',
    y='Oscillations detected \n (per Model fit)',
    hue='method',
    hue_order=['specparam', 'pyrasa'],
    palette='deep',
    ax=ax[2],
)
ax[2].set_xlabel('')
sns.despine()
f.tight_layout()

f.savefig('../results/bw_delta_sim.svg')

# %% Setting #5 - height_delta
sigs = []
h_list = [1, 1.5, 2]
for height in h_list:
    bw = 0.5
    freq = 10
    exp = -1.5
    peak_params = {'freq': freq, 'bw': bw, 'height': height}

    sig_ap = sim_powerlaw(n_seconds=n_secs, fs=fs, exponent=exp)
    sig = sim_peak_oscillation(sig_ap, fs=fs, **peak_params)

    f, psd = dsp.welch(sig, fs=fs, **kwargs_psd)
    df = pd.DataFrame({'Frequency (Hz)': f, 'Power': psd})
    df['height'] = height
    sigs.append(df)

df_1 = pd.concat(sigs)

# %%
f, ax = plt.subplots(figsize=(15, 5), ncols=3)

sns.lineplot(
    df_1.query('`Frequency (Hz)` < 50').query('`Frequency (Hz)` > 2'),
    legend=False,
    x='Frequency (Hz)',
    y='Power',
    hue='height',
    palette=pal,
    ax=ax[0],
)

ax[0].set_xscale('log')
ax[0].set_yscale('log')

df_ap, df_p = get_simulations('freq_height_delta')
df_ap = df_ap.reset_index()
df_ap['Exponent Δ'] = np.abs(df_ap['Exponent'] - np.abs(df_ap['sim_exp']))

n_runs = len(df_p['seed'].unique()) * len(df_p['sim_freq'].unique())
df_p = df_p.query('cf < 40').query('cf > 4').reset_index()
df_p.loc[df_p['method'] == 'pyrasa', 'pw'] = (df_p.loc[df_p['method'] == 'pyrasa', 'pw']) * 45
df_p['Power Δ'] = np.abs((df_p['sim_height'] - df_p['pw']))


sns.lineplot(df_p, legend=False, x='sim_height', y='pw', hue='method', palette='deep', ax=ax[1])
ax[1].set_xlabel('Simulated Peak Height')
ax[1].set_ylabel('Peak Height')

df_p2 = df_p[['cf', 'pw', 'bw', 'method']].dropna(axis=0)
osc_per_run = pd.DataFrame(df_p2.groupby('method').apply(len) / n_runs).reset_index()
osc_per_run.columns = ['method', 'Oscillations detected \n (per Model fit)']
sns.barplot(
    osc_per_run,
    x='method',
    y='Oscillations detected \n (per Model fit)',
    hue='method',
    hue_order=['specparam', 'pyrasa'],
    palette='deep',
    ax=ax[2],
)
ax[2].set_xlabel('')
sns.despine()

f.tight_layout()

f.savefig('../results/pw_delta_sim.svg')

# %%


# %%
plt.scatter(
    df_p.query('method == "specparam"')['pw'],  # *15,
    df_p.query('method == "specparam"')['sim_height'],
)


# %%
plt.scatter(
    df_p.query('method == "pyrasa"')['pw'],  # *15,
    df_p.query('method == "pyrasa"')['sim_height'],
)

# %%
sts.pearsonr(
    df_p.query('method == "pyrasa"').dropna(axis=1)['pw'],  # *15,
    df_p.query('method == "pyrasa"').dropna(axis=1)['sim_height'],
)


# %%
import scipy.signal as dsp

test_arr = np.array(
    [
        1,
        1,
        1,
        2,
        3,
        4,
        5,
        4,
        3,
        2,
        1,
        1,
    ]
)
dsp.find_peaks(test_arr, height=[0, 5], prominence=0.5)
# %%
height = 1
freq = 10
exp = -1.0
bw = 0.5
peak_params = {'freq': freq, 'bw': bw, 'height': height}
n_secs = 60
fs = 500
sig_ap = sim_powerlaw(n_seconds=n_secs, fs=fs, exponent=exp)
sig = sim_peak_oscillation(sig_ap, fs=fs, **peak_params)

# %%
freq, psd = dsp.welch(sig, fs=fs, nperseg=fs * 2, noverlap=fs)
psd.max()
# %%
plt.loglog(freq, psd)
# %%
dsp.find_peaks(
    psd,
    height=[0, 5],
    prominence=0.01,
)
# %%
from pyrasa import irasa
from pyrasa.utils.peak_utils import get_peak_params

# %%
freq, ap, p = irasa(sig, fs=fs, band=(1, 100), kwargs_psd={'nperseg': fs * 2, 'noverlap': fs})
# %%
plt.plot(freq, p[0, :])
# %%
dsp.find_peaks(p[0, :], height=[0, 1], prominence=0.05)
# %%
get_peak_params(p, freq)
# %%



import numpy as np
import pytest
import scipy.signal as dsp

from pyrasa import irasa
from pyrasa.utils.aperiodic_utils import compute_aperiodic_model
from pyrasa.utils.fit_funcs import AbstractFitFun

from .settings import EXPONENT, FS, HIGH_TOLERANCE, MIN_R2, TOLERANCE


# Test slope fitting functionality
@pytest.mark.parametrize('exponent', EXPONENT, scope='session')
@pytest.mark.parametrize('fs', FS, scope='session')
# @pytest.mark.xfail
def test_slope_fitting_fixed(fixed_aperiodic_signal, fs, exponent):
    f_range = [1, 100]
    # test whether recombining periodic and aperiodic spectrum is equivalent to the original spectrum
    freqs, psd = dsp.welch(fixed_aperiodic_signal, fs, nperseg=int(4 * fs))
    freq_logical = np.logical_and(freqs >= f_range[0], freqs <= f_range[1])
    freqs, psd = freqs[freq_logical], psd[freq_logical]

    # test whether we can reconstruct the exponent correctly
    aperiodic_fit_f = compute_aperiodic_model(psd, freqs, fit_func='fixed')
    assert pytest.approx(aperiodic_fit_f.aperiodic_params['Exponent'][0], abs=TOLERANCE) == np.abs(exponent)
    # test goodness of fit should be close to r_squared == 1 for linear model
    assert aperiodic_fit_f.gof['R2'][0] > MIN_R2

    # test if we can set fit bounds w/o error
    # _, _ = compute_slope(psd, freqs, fit_func='fixed', fit_bounds=[2, 50])

    # bic and aic for fixed model should be better if linear
    aperiodic_fit_k = compute_aperiodic_model(psd, freqs, fit_func='knee')
    # assert gof_k['AIC'][0] > gof['AIC'][0]
    assert aperiodic_fit_k.gof['BIC'][0] > aperiodic_fit_f.gof['BIC'][0]

    # test the effect of scaling
    aperiodic_fit_fs = compute_aperiodic_model(psd, freqs, fit_func='fixed', scale=True)
    assert np.isclose(aperiodic_fit_fs.aperiodic_params['Exponent'], aperiodic_fit_f.aperiodic_params['Exponent'])
    assert np.isclose(aperiodic_fit_fs.gof['R2'], aperiodic_fit_f.gof['R2'])


@pytest.mark.parametrize('exponent, fs', [(-1, 500)], scope='session')
def test_slope_fitting_settings(
    fixed_aperiodic_signal,
    exponent,
    fs,
):
    f_range = [0, 100]
    # test whether recombining periodic and aperiodic spectrum is equivalent to the original spectrum
    freqs, psd = dsp.welch(fixed_aperiodic_signal, fs, nperseg=int(4 * fs))
    freq_logical = np.logical_and(freqs >= f_range[0], freqs <= f_range[1])

    # match_txt = (
    #     'The first frequency appears to be 0 this will result in slope fitting problems. '
    #     + 'Frequencies will be evaluated starting from the next highest in Hz'
    # )
    # test bounds too low
    with pytest.raises(AssertionError):
        compute_aperiodic_model(psd[freq_logical], freqs[freq_logical], fit_func='fixed', fit_bounds=(0, 200))

    # test bounds too high
    with pytest.raises(AssertionError):
        compute_aperiodic_model(psd[freq_logical], freqs[freq_logical], fit_func='fixed', fit_bounds=(1, 1000))

    # test bounds correct
    compute_aperiodic_model(psd[freq_logical], freqs[freq_logical], fit_func='fixed', fit_bounds=(5, 40))

    # test for warning
    with pytest.warns(UserWarning):  # , match=match_txt):
        compute_aperiodic_model(psd[freq_logical], freqs[freq_logical], fit_func='fixed')

    # test misspecify string in fit_func
    with pytest.raises(ValueError):
        compute_aperiodic_model(psd[freq_logical], freqs[freq_logical], fit_func='incredible', fit_bounds=(5, 40))


# test custom slope fitting functions
@pytest.mark.parametrize('exponent, fs', [(-1, 500)], scope='session')
def test_custom_slope_fitting(
    fixed_aperiodic_signal,
    exponent,
    fs,
):
    f_range = [1.5, 100]
    # test whether recombining periodic and aperiodic spectrum is equivalent to the original spectrum
    freqs, psd = dsp.welch(fixed_aperiodic_signal, fs, nperseg=int(4 * fs))
    freq_logical = np.logical_and(freqs >= f_range[0], freqs <= f_range[1])
    psd, freqs = psd[freq_logical], freqs[freq_logical]

    class CustomFitFun(AbstractFitFun):
        def func(self, x: np.ndarray, a: float, b: float) -> np.ndarray:
            """
            Specparams fixed fitting function.
            Use this to model aperiodic activity without a spectral knee
            """
            y_hat = a + b * x

            return y_hat

        @property
        def curve_kwargs(self) -> dict[str, any]:
            aperiodic_nolog = self.aperiodic_spectrum
            off_guess = [aperiodic_nolog[0]]
            exp_guess = [
                np.abs(np.log10(aperiodic_nolog[0] / aperiodic_nolog[-1]) / np.log10(self.freq[-1] / self.freq[0]))
            ]
            return {
                'maxfev': 10_000,
                'ftol': 1e-5,
                'xtol': 1e-5,
                'gtol': 1e-5,
                'p0': np.array(off_guess + exp_guess),
                'bounds': ((-np.inf, -np.inf), (np.inf, np.inf)),
            }

    aperiodic_fit = compute_aperiodic_model(np.log10(psd), np.log10(freqs), fit_func=CustomFitFun)
    # add a high tolerance
    assert pytest.approx(np.abs(aperiodic_fit.aperiodic_params['b'][0]), abs=HIGH_TOLERANCE) == np.abs(exponent)

    class CustomFitFun(AbstractFitFun):
        log10_aperiodic = True
        log10_freq = True

        def func(self, x: np.ndarray, a: float, b: float) -> np.ndarray:
            """
            Specparams fixed fitting function.
            Use this to model aperiodic activity without a spectral knee
            """
            y_hat = a + b * x

            return y_hat

    irasa_spectrum = irasa(fixed_aperiodic_signal, fs, f_range, psd_kwargs={'nperseg': 4 * fs})
    aperiodic_fit = irasa_spectrum.fit_aperiodic_model(fit_func=CustomFitFun)

    # add a high tolerance
    assert pytest.approx(np.abs(aperiodic_fit.aperiodic_params['b'][0]), abs=HIGH_TOLERANCE) == np.abs(exponent)



# %%
import numpy as np
from cluster_jobs.analyze_eo_ec import eo_ec_analysis
from plus_slurm import JobCluster, PermuteArgument


# % get jobcluster
job_cluster = JobCluster(
    required_ram='2G', request_cpus=2, request_time=30, python_bin='/home/schmidtfa/miniconda3/envs/pyrasa/bin/python'
)


sids = [str(i) for i in np.arange(2, 12).tolist()]

# %%
job_cluster.add_job(eo_ec_analysis, subject_id=PermuteArgument(sids), condition=PermuteArgument(['EO', 'EC']))

# % submit...
job_cluster.submit(do_submit=True)
# %%



#%%
from neurodsp.sim import sim_oscillation
import matplotlib.pyplot as plt
from pyrasa.utils.peak_utils import get_peak_params
import scipy.signal as dsp
import numpy as np
# %%
f_range = [1, 250]
n_secs = 5*60
fs = 200
osc_freq = 98
f_range4plot = [80, 110]

ts = sim_oscillation(n_seconds=n_secs, fs=fs, freq=osc_freq)
plt.plot(ts[:100])

#%%
freqs, psd = dsp.welch(ts, fs, nperseg=int(4 * fs))
plt.plot(freqs, psd)

freq_logical = np.logical_and(freqs >= f_range[0], freqs <= f_range[1])
freqs, psd = freqs[freq_logical], psd[freq_logical]
# %%
pe_params = get_peak_params(psd[np.newaxis, :], freqs, min_peak_height=0.1)
pe_params
# %%

freq_logical = np.logical_and(freqs >= f_range4plot[0], freqs <= f_range4plot[1])
plt.plot(freqs[freq_logical], psd[freq_logical])
# %%



#%%
from neurodsp.sim import set_random_seed
from neurodsp.sim import sim_powerlaw, sim_oscillation
from neurodsp.utils import create_times
from neurodsp.plts import plot_timefrequency#

from neurodsp.timefrequency import compute_wavelet_transform
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns

sns.set_style('ticks')
sns.set_context('talk')

set_random_seed(42)

from pyrasa.irasa import irasa_sprint
# %%
# Set some general settings, to be used across all simulations
fs = 500
n_seconds = 15
duration=4
overlap=0.5

# Create a times vector for the simulations
times = create_times(n_seconds, fs)

#%%
alpha = sim_oscillation(n_seconds=.5, fs=fs, freq=10)
no_alpha = np.zeros(len(alpha))
beta = sim_oscillation(n_seconds=.5, fs=fs, freq=25)
no_beta = np.zeros(len(beta))

exp_1 = sim_powerlaw(n_seconds=2.5, fs=fs, exponent=-1)
exp_2 = sim_powerlaw(n_seconds=2.5, fs=fs, exponent=-2)

# %%
alphas = np.concatenate([no_alpha, alpha, no_alpha, alpha, no_alpha])
betas = np.concatenate([beta, no_beta, beta, no_beta, beta])

sim_ts = np.concatenate([exp_1 + alphas, 
                         exp_1 + alphas + betas, 
                         exp_1 + betas, 
                         exp_2 + alphas, 
                         exp_2 + alphas + betas, 
                         exp_2 + betas, ])
# %%
plt.plot(times, sim_ts)
# %%
freqs = np.arange(1, 50, 0.5)

#%%
mwt = compute_wavelet_transform(sim_ts, fs=fs, freqs=freqs, n_cycles=11,)
#%%
plot_timefrequency(times, freqs, mwt, vmin=0)

# %%
sgramm_ap, sgramm_p, freqs_ir, times_ir = irasa_sprint(sim_ts[np.newaxis, :], fs=fs,
                                                       band=(1, 100),
                                                        
                                                       freq_res=0.5,
                                                       smooth=False, 
                                                       n_avgs=[3, 7, 11])



# %% now extract the aperiodic features
from pyrasa.utils.aperiodic_utils import compute_slope_sprint

df_aps, df_gof = compute_slope_sprint(sgramm_ap, freqs=freqs_ir, 
                                      times=times_ir, fit_func='fixed')


plot_timefrequency(times_ir, freqs_ir, np.squeeze(sgramm_ap), vmin=0)


#%%
ave = df_gof['r_squared'].mean()
std = df_gof['r_squared'].std()
print(f'Average goodness of fit {ave} and deviation is {std}')


#%%
ave = df_gof.query('time < 7')['r_squared'].mean()
std = df_gof.query('time < 7')['r_squared'].std()
print(f'Average goodness of fit {ave} and deviation is {std}')


#%%
ave = df_gof.query('time > 7')['r_squared'].mean()
std = df_gof.query('time > 7')['r_squared'].std()
print(f'Average goodness of fit {ave} and deviation is {std}')


# %%
f, ax = plt.subplots(nrows=2, figsize=(8, 4))

ax[0].plot(df_aps['time'], df_aps['Exponent'])
ax[0].set_ylabel('Exponent')
ax[0].set_xlabel('time (s)')
ax[1].plot(df_aps['time'], df_gof['r_squared'])
ax[1].set_ylabel('R2')
ax[1].set_xlabel('time (s)')

f.tight_layout()

# %%
# %%
from pyrasa.utils.peak_utils import get_peak_params_sprint

df_peaks = get_peak_params_sprint(sgramm_p, 
                                  freqs=freqs_ir, 
                                  times=times_ir,
                                  min_peak_height=0.01)


plot_timefrequency(times_ir, freqs_ir, np.squeeze(sgramm_p), vmin=0)

# %% Plot peak results
f, ax = plt.subplots(nrows=3, figsize=(12, 4))

for ix, cur_key in enumerate(['cf', 'pw', 'bw']):

    ax[ix].plot(df_peaks['time'], df_peaks[cur_key])
    ax[ix].set_ylabel(cur_key)
    ax[ix].set_xlabel('time (s)')

#%%


# %%
import numpy as np
from cluster_jobs.super_simulation import super_simulation
from plus_slurm import JobCluster, PermuteArgument


# % get jobcluster
job_cluster = JobCluster(
    required_ram='2G', request_cpus=1, request_time=30, python_bin='/home/schmidtfa/miniconda3/envs/pyrasa/bin/python'
)

# % put in jobs...
sim_kinds = ['fixed', 'knee', 'burst', 'broad_fixed', 'broad_knee']
EXPS = np.round(np.arange(-3, 0.5, 0.5), 1).tolist()
KNEES = [25, 100, 400, 900, 1600]

FREQS = np.arange(5, 36, 1).tolist()
POWERS = np.round(np.arange(0, 2.0, 0.2), 1).tolist()
BWS = np.round(np.arange(0.5, 3.5, 0.5), 1).tolist()

# Burst related parameters
BPROBS = np.round(np.arange(0.2, 0.8, 0.2), 1).tolist()
# len(BPROBS) * len(BWS) * len(POWERS) * len(FREQS) * len(KNEES) * len(EXPS)
# %%
job_cluster.add_job(
    super_simulation,
    subject_id='broad_knee',  # PermuteArgument(sim_kinds),
    exp=PermuteArgument(EXPS),
    knee=PermuteArgument(KNEES),
    freq=PermuteArgument(FREQS),
    bw=PermuteArgument(BWS),
    height=PermuteArgument(POWERS),
    # b_proba = PermuteArgument(BPROBS),
)

# % submit...
job_cluster.submit(do_submit=True)
# %%



#%%
import numpy as np
import scipy.signal as dsp
from pyrasa.utils.peak_utils import get_peak_params
from neurodsp.sim import sim_powerlaw


fs = 500

sig = sim_powerlaw(n_seconds=60, fs=fs, exponent=-1)
f_range = [0, 100]
#%% test whether recombining periodic and aperiodic spectrum is equivalent to the original spectrum
freqs, psd = dsp.welch(sig, fs, nperseg=int(4 * fs))
freq_logical = np.logical_and(freqs >= f_range[0], freqs <= f_range[1])
get_peak_params(psd[freq_logical], freqs[freq_logical], min_peak_height=10)
# %%



import pytest
from neurodsp.utils.sim import set_random_seed
from scipy.optimize import OptimizeWarning

from pyrasa.irasa_mne import irasa_epochs, irasa_raw

set_random_seed(42)


@pytest.mark.filterwarnings('ignore:RuntimeWarning')
@pytest.mark.filterwarnings('ignore:UserWarning')
def test_mne(gen_mne_data_raw):
    mne_data, epochs = gen_mne_data_raw

    # test raw
    irasa_raw_result = irasa_raw(mne_data, band=(0.25, 50), duration=2, hset_info=(1.0, 2.0, 0.05))
    with pytest.warns(OptimizeWarning):
        irasa_raw_result.aperiodic.fit_aperiodic_model(fit_func='fixed')
        irasa_raw_result.periodic.get_peaks(smoothing_window=2)

    # test epochs
    irasa_epoched_result = irasa_epochs(epochs, band=(0.5, 50), hset_info=(1.0, 2.0, 0.05))
    irasa_epoched_result.aperiodic.fit_aperiodic_model(fit_func='fixed', scale=True)
    irasa_epoched_result.periodic.get_peaks(smoothing_window=2)



# %%

import numpy as np

N_SECONDS = 60
FS = [500, 750, 1000]

OSC_FREQ = [5, 10, 20]
MANY_OSC_FREQ = np.arange(2, 30, 1)
# through a bug in neurodsp #223 there is a systematic error in generating spectra with high frequency peaks
# where the actual simulated center frequency is biased.
# This bias increases as the simulated center frequency gets closer to the sampling rate used in the simulation.
# If you want to play around with this check: simulations/notebooks/sim_peak_tests.py
EXPONENT = [-1, -1.5, -2.0]
KNEE_FREQ = 15
KNEE_FREQS = [10, 15, 25]
# There seems to be a higher error in knee fits for knee and exponent estimates when
# the difference in pre and post knee exponent is low. This kinda makes sense
# TODO: Test this systematically -> see whether this is an issue with irasa or slope fitting in general
EXP_KNEE_COMBO = [
    (-1.0, 10.0),
    (-1.0, 15.0),
    (-1.0, 25.0),
    (-1.5, 125.0),
    (-1.5, 32.0),
    (-1.5, 58.0),
    (-2.0, 100.0),
    (-2.0, 225.0),
    (-2.0, 625.0),
]  # we test exp + knee combined as both relate to each other
TOLERANCE = 0.3  # 0.15
SPRINT_TOLERANCE = 0.4
HIGH_TOLERANCE = 0.5

KNEE_TOLERANCE = 5
MIN_R2 = 0.8  # seems like a sensible minimum
MIN_R2_SPRINT = 0.7
MIN_CORR_PSD_CMB = 0.99



"""Classes used to model aperiodic spectra"""

import abc
import inspect
from collections.abc import Callable
from typing import Any, ClassVar, no_type_check

import numpy as np
import pandas as pd
from attrs import define
from scipy.optimize import curve_fit


def _get_args(f: Callable) -> list:
    """
    Extracts the argument names from a function, excluding the first two.

    Parameters
    ----------
    f : Callable
        The function or method from which to extract argument names.

    Returns
    -------
    list
        A list of argument names, excluding the first two.
    """

    return inspect.getfullargspec(f)[0][2:]


def _get_gof(psd: np.ndarray, psd_pred: np.ndarray, k: int, fit_type: str) -> pd.DataFrame:
    """
    Calculate the goodness of fit metrics for a given model prediction against
    actual aperiodic power spectral density (PSD) data.

    This function computes several statistics to evaluate how well the predicted PSD values
    match the observed PSD values. The metrics include Mean Squared Error (MSE), R-squared (R²),
    Bayesian Information Criterion (BIC), and Akaike Information Criterion (AIC).

    Parameters
    ----------
    psd : np.ndarray
        The observed power spectral density values.
    psd_pred : np.ndarray
        The predicted power spectral density values from the model.
    k : int
        The number of parameters in the curve fitting function used to predict the `psd`.
    fit_type : str
        A description or label for the type of fit/model used, which will be included in the output DataFrame.

    Returns
    -------
    pd.DataFrame
        A DataFrame containing the goodness of fit metrics:
        - 'mse': Mean Squared Error
        - 'r_squared': R-squared value
        - 'BIC': Bayesian Information Criterion
        - 'AIC': Akaike Information Criterion
        - 'fit_type': The type of fit/model used (provided as input)

    Notes
    -----
    - BIC and AIC calculations currently assume Ordinary Least Squares (OLS) regression.

    References
    ----------
    For further details on BIC and AIC, see: https://machinelearningmastery.com/probabilistic-model-selection-measures/
    """

    residuals = psd - psd_pred
    ss_res = np.sum(residuals**2)
    ss_tot = np.sum((psd - np.mean(psd)) ** 2)

    mse = np.mean(residuals**2)
    n = len(psd)

    # https://robjhyndman.com/hyndsight/lm_aic.html
    # c is in practice sometimes dropped. Only relevant when comparing models with different n
    # c = n + n * np.log(2 * np.pi)
    # aic = 2 * k + n * np.log(mse) + c #real
    aic = 2 * k + np.log(n) * np.log(mse)  # + c
    # aic = 2 * k + n * mse
    # according to Sclove 1987 only difference between BIC and AIC
    # is that BIC uses log(n) * k instead of 2 * k
    # bic = np.log(n) * k + n * np.log(mse) + c #real
    bic = np.log(n) * k + np.log(n) * np.log(mse)  # + c
    # bic = np.log(n) * k + n * mse
    # Sclove 1987 also hints at sample size adjusted bic
    an = (n + 2) / 24  # defined in Rissanen 1978 based on minimum-bit representation of a signal
    # abic -> https://www.ncbi.nlm.nih.gov/pmc/articles/PMC7299313/
    abic = np.log(an) * k + np.log(n) * np.log(mse)

    r2 = 1 - (ss_res / ss_tot)
    r2_adj = 1 - (((1 - r2) * (n - 1)) / (n - k - 1))

    gof = pd.DataFrame({'mse': mse, 'R2': r2, 'R2_adj.': r2_adj, 'BIC': bic, 'BIC_adj.': abic, 'AIC': aic}, index=[0])
    gof['fit_type'] = fit_type
    return gof


@define
class AbstractFitFun(abc.ABC):
    """
    Abstract base class for fitting functions used to model aperiodic spectra.

    This class provides a framework for defining and fitting models to aperiodic spectra.
    It handles common functionality required for fitting a model, such as scaling and goodness-of-fit
    computation. Subclasses should implement the `func` method to define the specific fitting function
    used for curve fitting.

    Attributes
    ----------
    freq : np.ndarray
        The frequency values associated with the aperiodic spectrum data.
    aperiodic_spectrum : np.ndarray
        The aperiodic spectrum data to which the model will be fit.
    scale_factor : int | float
        A scaling factor used to adjust the fit results.
    label : ClassVar[str]
        A label to identify the type of fit or model used. Default is 'custom'.
    log10_aperiodic : ClassVar[bool]
        If True, the aperiodic spectrum values will be transformed using log10. Default is False.
    log10_freq : ClassVar[bool]
        If True, the frequency values will be transformed using log10. Default is False.

    Methods
    -------
    __attrs_post_init__()
        Post-initialization method to apply log10 transformations if specified.
    func(x: np.ndarray, *args: float) -> np.ndarray
        Abstract method to define the model function. Must be implemented by subclasses
        and should be applicable to scipy.optimize.curve_fit.
    curve_kwargs() -> dict[str, Any]
        Returns keyword arguments for the curve fitting process.
    add_infos_to_df(df_params: pd.DataFrame) -> pd.DataFrame
        Method to add additional information to the parameters DataFrame. Can be overridden by subclasses.
    handle_scaling(df_params: pd.DataFrame, scale_factor: float) -> pd.DataFrame
        Adjusts the parameters DataFrame based on the scaling factor. Can be overridden by subclasses.
    fit_func() -> tuple[pd.DataFrame, pd.DataFrame]
        Fits the model to the data and returns DataFrames containing the model parameters and goodness-of-fit metrics.

    Notes
    -----
    - Subclasses must implement the `func` method to define the model's functional form.
    - The `curve_kwargs` method can be overridden to customize curve fitting options.
    - The `add_infos_to_df` and `handle_scaling` methods are intended to be overridden if additional
      functionality or specific scaling behavior is required.

    References
    ----------
    For details on goodness-of-fit metrics and their calculations, see the documentation for `_get_gof`.
    """

    freq: np.ndarray
    aperiodic_spectrum: np.ndarray
    scale_factor: int | float
    label: ClassVar[str] = 'custom'
    log10_aperiodic: ClassVar[bool] = False
    log10_freq: ClassVar[bool] = False

    def __attrs_post_init__(self) -> None:
        if self.log10_aperiodic:
            self.aperiodic_spectrum = np.log10(self.aperiodic_spectrum)
        if self.log10_freq:
            self.freq = np.log10(self.freq)

    @abc.abstractmethod
    @no_type_check
    def func(self, x: np.ndarray, *args: float) -> np.ndarray:
        pass

    @property
    def curve_kwargs(self) -> dict[str, Any]:
        return {
            'maxfev': 10_000,
            'ftol': 1e-5,
            'xtol': 1e-5,
            'gtol': 1e-5,
        }

    def add_infos_to_df(self, df_params: pd.DataFrame) -> pd.DataFrame:
        return df_params

    def handle_scaling(self, df_params: pd.DataFrame, scale_factor: float) -> pd.DataFrame:
        if 'Offset' in df_params.columns:
            df_params['Offset'] /= scale_factor
        elif scale_factor != 1.0:
            raise ValueError('Scale Factor not handled. You need to overwrite the handle_scaling method.')
        return df_params

    def fit_func(self) -> tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        curve_kwargs = self.curve_kwargs
        p, _ = curve_fit(self.func, self.freq, self.aperiodic_spectrum, **curve_kwargs)

        my_args = _get_args(self.func)
        df_params = pd.DataFrame(dict(zip(my_args, p)), index=[0])
        df_params['fit_type'] = self.label

        pred = self.func(self.freq, *p)
        df_gof = _get_gof(self.aperiodic_spectrum, pred, len(p), self.label)

        df_params = self.add_infos_to_df(df_params)
        df_params = self.handle_scaling(df_params, scale_factor=self.scale_factor)

        freq = self.freq.copy()
        if self.log10_aperiodic:
            pred = 10**pred
        if self.log10_freq:
            freq = 10**freq
        df_pred = pd.DataFrame({'Frequency (Hz)': freq, 'aperiodic_model': pred})
        df_pred['fit_type'] = self.label

        return df_params, df_gof, df_pred


class FixedFitFun(AbstractFitFun):
    """
    A model for fitting aperiodic activity in power spectra.

    The `FixedFitFun` class extends `AbstractFitFun` to model aperiodic activity in power spectra
    using a fixed function that does not include a spectral knee. This model is suitable for
    cases where the aperiodic component of the spectrum follows a consistent slope across
    the entire frequency range.

    Attributes
    ----------
    label : str
        A label to identify this fitting model. Default is 'fixed'.
    log10_aperiodic : bool
        Indicates whether to log-transform the aperiodic spectrum. Default is True.

    Methods
    -------
    func(x: np.ndarray, Offset: float, Exponent: float) -> np.ndarray
        Defines the model function for aperiodic activity without a spectral knee.

    curve_kwargs() -> dict[str, Any]
        Generates initial guess parameters and other keyword arguments for curve fitting.
    """

    label = 'fixed'
    log10_aperiodic = True

    def func(self, x: np.ndarray, Offset: float, Exponent: float) -> np.ndarray:  # noqa N803
        """
        Specparams fixed fitting function.
        Use this to model aperiodic activity without a spectral knee
        """
        y_hat = Offset - np.log10(x**Exponent)

        return y_hat

    @property
    def curve_kwargs(self) -> dict[str, Any]:
        aperiodic_nolog = 10**self.aperiodic_spectrum
        off_guess = [aperiodic_nolog[0]]
        exp_guess = [
            np.abs(np.log10(aperiodic_nolog[0] / aperiodic_nolog[-1]) / np.log10(self.freq[-1] / self.freq[0]))
        ]
        return {
            'maxfev': 10_000,
            'ftol': 1e-5,
            'xtol': 1e-5,
            'gtol': 1e-5,
            'p0': np.array(off_guess + exp_guess),
            'bounds': ((-np.inf, -np.inf), (np.inf, np.inf)),
        }


class KneeFitFun(AbstractFitFun):
    """
    A model for fitting aperiodic activity in power spectra with a spectral knee.

    The `KneeFitFun` class extends `AbstractFitFun` to model aperiodic activity in power spectra
    using a function that includes a spectral knee. This model is particularly useful for
    cases where the aperiodic component of the spectrum has a break or knee, representing
    a transition between two different spectral slopes.

    Attributes
    ----------
    label : str
        A label to identify this fitting model. Default is 'knee'.
    log10_aperiodic : bool
        Indicates whether to log-transform the aperiodic spectrum. Default is True.

    Methods
    -------
    func(x: np.ndarray, Offset: float, Knee: float, Exponent_1: float, Exponent_2: float) -> np.ndarray
        Defines the model function for aperiodic activity with a spectral knee and pre-knee slope.

    add_infos_to_df(df_params: pd.DataFrame) -> pd.DataFrame
        Adds calculated knee frequency to the DataFrame of fit parameters.

    curve_kwargs() -> dict[str, Any]
        Generates initial guess parameters and other keyword arguments for curve fitting.
    """

    label = 'knee'
    log10_aperiodic = True

    def func(
        self,
        x: np.ndarray,
        Offset: float,  # noqa N803
        Knee: float,  # noqa N803
        Exponent_1: float,  # noqa N803
        Exponent_2: float,  # noqa N803
    ) -> np.ndarray:
        """
        Model aperiodic activity with a spectral knee and a pre-knee slope.
        Use this to model aperiodic activity with a spectral knee
        """
        y_hat = Offset - np.log10(x**Exponent_1 * (Knee + x**Exponent_2))

        return y_hat

    def add_infos_to_df(self, df_params: pd.DataFrame) -> pd.DataFrame:
        df_params['Knee Frequency (Hz)'] = df_params['Knee'] ** (
            1.0 / (2 * df_params['Exponent_1'] + df_params['Exponent_2'])
        )
        df_params['tau'] = 1.0 / (2 * np.pi * df_params['Knee Frequency (Hz)'])
        return df_params

    @property
    def curve_kwargs(self) -> dict[str, Any]:
        aperiodic_nolog = 10**self.aperiodic_spectrum
        off_guess = [aperiodic_nolog[0]]
        exp_guess = [
            np.abs(np.log10(aperiodic_nolog[0] / aperiodic_nolog[-1]) / np.log10(self.freq[-1] / self.freq[0]))
        ]
        cumsum_psd = np.cumsum(aperiodic_nolog)
        half_pw_freq = self.freq[np.abs(cumsum_psd - (0.5 * cumsum_psd[-1])).argmin()]
        # make the knee guess the point where we have half the power in the spectrum seems plausible to me
        knee_guess = [half_pw_freq ** (exp_guess[0] + exp_guess[0])]

        return {
            'maxfev': 10_000,
            'ftol': 1e-5,
            'xtol': 1e-5,
            'gtol': 1e-5,
            'p0': np.array(off_guess + knee_guess + exp_guess + exp_guess),
            'bounds': ((0, 0, 0, 0), (np.inf, np.inf, np.inf, np.inf)),
        }



# %%
import sys
from neurodsp.sim import sim_combined
import numpy as np
import seaborn as sns

sns.set_style('ticks')
sns.set_context('paper')
sys.path.append('../')


import matplotlib as mpl

new_rc_params = {'text.usetex': False, 'svg.fonttype': 'none'}
mpl.rcParams.update(new_rc_params)

import mne
from mne.datasets import sample


from pyrasa.irasa_mne import irasa_raw, irasa_epochs

# %% now lets check the mne python implementation


data_path = sample.data_path()
subjects_dir = data_path / 'subjects'
subject = 'sample'

meg_path = data_path / 'MEG' / 'sample'
raw_fname = meg_path / 'sample_audvis_raw.fif'
fwd_fname = meg_path / 'sample_audvis-meg-eeg-oct-6-fwd.fif'
fwd = mne.read_forward_solution(fwd_fname)
src = fwd['src']


raw = mne.io.read_raw_fif(raw_fname)
picks = mne.pick_types(raw.info, meg='mag', eeg=False, stim=False, eog=False, exclude='bads')
raw.pick(picks)

# %% lets simulate 2 different signals
info = raw.info.copy()
tstep = 1.0 / info['sfreq']

alpha = sim_combined(
    60,
    info['sfreq'],
    components={
        'sim_oscillation': {'freq': 10},
        'sim_powerlaw': {
            'exponent': -1.5,
        },
    },
)

beta = sim_combined(
    60,
    info['sfreq'],
    components={
        'sim_oscillation': {'freq': 20},
        'sim_powerlaw': {
            'exponent': -1.5,
        },
    },
)


# %% now lets simulate some data
# TODO: put in a helper function


def simulate_raw_(signal, scaling_fator, region, subject, info, subjects_dir):
    """Shorthand function to simulate a dipole"""

    selected_label = mne.read_labels_from_annot(
        subject,
        regexp=region,
        subjects_dir=subjects_dir,
    )[0]
    location = 'center'  # Use the center of the region as a seed.
    extent = 10.0  # Extent in mm of the region.
    label = mne.label.select_sources(
        subject, selected_label, location=location, extent=extent, subjects_dir=subjects_dir
    )

    # this is the source ts
    source_time_series = signal * scaling_fator

    n_events = 1
    events = np.zeros((n_events, 3), int)
    events[:, 0] = np.arange(n_events)  # Events sample.
    events[:, 2] = 1  # All events have the sample id.

    source_simulator = mne.simulation.SourceSimulator(src, tstep=tstep)
    source_simulator.add_data(label, source_time_series, events)

    raw = mne.simulation.simulate_raw(info, source_simulator, forward=fwd)
    # cov = mne.make_ad_hoc_cov(raw.info)
    # mne.simulation.add_noise(raw, cov, iir_filter=[0.2, -0.2, 0.04])

    return raw


# %% simulate alpha + beta and combine
raw_alpha = simulate_raw_(alpha, 5e-9, 'caudalmiddlefrontal-rh', subject, info, subjects_dir)

raw_beta = simulate_raw_(beta, 10e-10, 'caudalmiddlefrontal-lh', subject, info, subjects_dir)

raw_data = raw_alpha.get_data() + raw_beta.get_data()

raw = mne.io.RawArray(raw_data, info)

picks = mne.pick_types(raw.info, meg='mag', eeg=False, stim=False, eog=False, exclude='bads')
raw.pick(picks)

# %%
aperiodic_mne, periodic_mne = irasa_raw(raw, band=(0.25, 50), duration=2, hset_info=(1.0, 2.0, 0.05), as_array=False)
# %%
df_slopes, df_gof = aperiodic_mne.get_slopes()

# %%
df_slopes.head()
# %%
df_gof.head()

# %%

peak_df = periodic_mne.get_peaks()

peak_df.head()

periodic_mne.plot()

# %% now lets check-out the events
event_id = {
    'Auditory/Left': 1,
    'Auditory/Right': 2,
    'Visual/Left': 3,
    'Visual/Right': 4,
}
tmin = -0.2
tmax = 0.5

# Load real data as the template
event_fname = meg_path / 'sample_audvis_filt-0-40_raw-eve.fif'
events = mne.read_events(event_fname)


epochs = mne.Epochs(
    raw,
    events,
    event_id,
    tmin,
    tmax,
    # picks=picks,
    baseline=None,
    preload=True,
    verbose=False,
)

# %%

aperiodic, periodic = irasa_epochs(epochs, band=(0.5, 50), hset_info=(1.0, 2.0, 0.05), as_array=False)

# %%
periodic.get_peaks(polyorder=0)

# %%
aperiodic.get_slopes()

# %%
periodic['Auditory/Left'].average().plot_topomap(dB=False)
# %%
periodic['Auditory/Right'].average().plot_topomap(dB=False)
# %%
periodic['Visual/Left'].average().plot_topomap(dB=False)
# %%
periodic['Visual/Right'].average().plot_topomap(dB=False)


# %%
aperiodic['Auditory/Left'].average().plot_topomap(dB=False)
# %%
aperiodic['Auditory/Right'].average().plot_topomap(dB=False)
# %%
aperiodic['Visual/Left'].average().plot_topomap(dB=False)
# %%
aperiodic['Visual/Right'].average().plot_topomap(dB=False)

# %%



"""Interface to use the IRASA algorithm with MNE objects."""

import mne
import numpy as np

from pyrasa.irasa import irasa
from pyrasa.irasa_mne.mne_objs import (
    AperiodicEpochsSpectrum,
    AperiodicSpectrumArray,
    IrasaEpoched,
    IrasaRaw,
    PeriodicEpochsSpectrum,
    PeriodicSpectrumArray,
)


def irasa_raw(
    data: mne.io.Raw,
    band: tuple[float, float] = (1.0, 100.0),
    duration: float | None = None,
    overlap: float | int = 50,
    hset_info: tuple[float, float, float] = (1.05, 2.0, 0.05),
) -> IrasaRaw:
    """
    Separate aperiodic from periodic power spectra using the IRASA algorithm.

    This function applies the Irregular Resampling Auto-Spectral Analysis (IRASA) algorithm
    as described by Wen & Liu (2016) to decompose the power spectrum of neurophysiological
    signals into aperiodic (fractal) and periodic (oscillatory) components. This function is
    essentially a wrapper function for `pyrasa.irasa`

    Parameters
    ----------
    data : mne.io.Raw
        The time-series data from which the aperiodic and periodic power spectra are extracted.
        This should be an instance of `mne.io.Raw`. The function will automatically extract
        relevant parameters such as sampling frequency (`sfreq`) and filtering settings from the `mne` object
        to make sure the model is specified correctly.
    band : tuple of (float, float), optional, default: (1.0, 100.0)
        A tuple specifying the lower and upper bounds of the frequency range (in Hz) used
        for extracting the aperiodic and periodic spectra.
    duration : float, required
        The duration (in seconds) of each segment used to calculate the power spectral density (PSD).
        This must be less than the total duration of the data.
    overlap : float or int, optional, default: 50
        The overlap between segments, specified as a percentage (0-100).
    hset_info : tuple of (float, float, float), optional, default: (1.05, 2.0, 0.05)
        Contains the range of up/downsampling factors used in the IRASA algorithm.
        This should be a tuple specifying the (min, max, step) values for the resampling.

    Returns
    -------
    IrasaRaw
        A custom object containing the separated aperiodic and periodic components of the data:
        - `periodic`: An instance of `PeriodicSpectrumArray`, which includes the periodic
          (oscillatory) component of the power spectrum.
        - `aperiodic`: An instance of `AperiodicSpectrumArray`, which includes the aperiodic
          (fractal) component of the power spectrum.

    References
    ----------
    Wen, H., & Liu, Z. (2016). Separating Fractal and Oscillatory Components in the Power Spectrum
    of Neurophysiological Signal. Brain Topography, 29(1), 13–26.
    https://doi.org/10.1007/s10548-015-0448-0

    Notes
    -----
    - Ensure that `data` does not contain any bad channels (`data.info['bads']` should be empty),
      as this could affect the results.
    - The overlap percentage should be carefully chosen to balance between segment independence
      and sufficient data for analysis. A value between 0 and 100% is valid.
    - The function will raise assertions if the input parameters are not consistent with the
      expected formats or if the provided `duration` exceeds the length of the data.

    """

    # set parameters & safety checks
    # ensure that input data is in the right format
    assert isinstance(data, mne.io.BaseRaw), 'Data should be of type mne.BaseRaw'
    assert data.info['bads'] == [], (
        'Data should not contain bad channels ' 'as this might mess up the creation of the returned data structure'
    )

    # extract relevant info from mne object
    info = data.info.copy()
    fs = data.info['sfreq']
    data_array = data.get_data()

    overlap /= 100
    assert isinstance(duration, int | float), 'You need to set the duration of your time window in seconds'
    assert data_array.shape[1] > int(fs * duration), 'The duration for each segment cant be longer than the actual data'
    assert np.logical_and(
        overlap < 1, overlap > 0
    ), 'The overlap between segments cant be larger than 100% or less than 0%'

    nfft = 2 ** (np.ceil(np.log2(int(fs * duration * np.max(hset_info)))))
    kwargs_psd = {
        'window': 'hann',
        'average': 'median',
        'nfft': nfft,
        'nperseg': int(fs * duration),
        'noverlap': int(fs * duration * overlap),
    }

    irasa_spectrum = irasa(
        data_array,
        fs=fs,
        band=band,
        filter_settings=(data.info['highpass'], data.info['lowpass']),
        hset_info=hset_info,
        psd_kwargs=kwargs_psd,
    )

    return IrasaRaw(
        periodic=PeriodicSpectrumArray(irasa_spectrum.periodic, info, freqs=irasa_spectrum.freqs),
        aperiodic=AperiodicSpectrumArray(irasa_spectrum.aperiodic, info, freqs=irasa_spectrum.freqs),
    )


def irasa_epochs(
    data: mne.Epochs,
    band: tuple[float, float] = (1.0, 100.0),
    hset_info: tuple[float, float, float] = (1.05, 2.0, 0.05),
) -> IrasaEpoched:
    """
    Separate aperiodic from periodic power spectra using the IRASA algorithm for Epochs data.

    This function applies the Irregular Resampling Auto-Spectral Analysis (IRASA) algorithm
    as described by Wen & Liu (2016) to decompose the power spectrum of neurophysiological
    signals into aperiodic (fractal) and periodic (oscillatory) components. It is specifically
    designed for time-series data in `mne.Epochs` format, making it suitable for event-related
    EEG/MEG analyses.

    Parameters
    ----------
    data : mne.Epochs
        The time-series data used to extract aperiodic and periodic power spectra.
        This should be an instance of `mne.Epochs`.
    band : tuple of (float, float), optional, default: (1.0, 100.0)
        A tuple specifying the lower and upper bounds of the frequency range (in Hz) used
        for extracting the aperiodic and periodic spectra.
    hset_info : tuple of (float, float, float), optional, default: (1.05, 2.0, 0.05)
        Contains the range of up/downsampling factors used in the IRASA algorithm.
        This should be a tuple specifying the (min, max, step) values for the resampling.

    Returns
    -------
    aperiodic : AperiodicEpochsSpectrum
        The aperiodic component of the data as an `AperiodicEpochsSpectrum` object.
    periodic : PeriodicEpochsSpectrum
        The periodic component of the data as a `PeriodicEpochsSpectrum` object.

    References
    ----------
    Wen, H., & Liu, Z. (2016). Separating Fractal and Oscillatory Components in the Power Spectrum
    of Neurophysiological Signal. Brain Topography, 29(1), 13–26.
    https://doi.org/10.1007/s10548-015-0448-0

    """

    # set parameters & safety checks
    # ensure that input data is in the right format
    assert isinstance(data, mne.BaseEpochs), 'Data should be of type mne.BaseEpochs'
    assert (
        data.info['bads'] == []
    ), 'Data should not contain bad channels as this might mess up the creation of the returned data structure'

    info = data.info.copy()
    fs = data.info['sfreq']
    events = data.events
    event_ids = data.event_id

    data_array = data.get_data(copy=True)

    # TODO: check if hset.max() is really max
    nfft = 2 ** (np.ceil(np.log2(int(data_array.shape[2] * np.max(hset_info)))))

    # TODO: does zero padding make sense?
    kwargs_psd = {
        'window': 'hann',
        'nperseg': None,
        'nfft': nfft,
        'noverlap': 0,
    }

    # Do the actual IRASA stuff..
    psd_list_aperiodic, psd_list_periodic = [], []
    for epoch in data_array:
        irasa_spectrum = irasa(
            epoch,
            fs=fs,
            band=band,
            filter_settings=(data.info['highpass'], data.info['lowpass']),
            hset_info=hset_info,
            psd_kwargs=kwargs_psd,
        )
        psd_list_aperiodic.append(irasa_spectrum.aperiodic.copy())
        psd_list_periodic.append(irasa_spectrum.periodic.copy())

    psds_aperiodic = np.array(psd_list_aperiodic)
    psds_periodic = np.array(psd_list_periodic)

    return IrasaEpoched(
        periodic=PeriodicEpochsSpectrum(
            psds_periodic, info, freqs=irasa_spectrum.freqs, events=events, event_id=event_ids
        ),
        aperiodic=AperiodicEpochsSpectrum(
            psds_aperiodic, info, freqs=irasa_spectrum.freqs, events=events, event_id=event_ids
        ),
    )



import numpy as np
import scipy.signal as dsp
import fractions


# %%cohrasa
def cohrasa(x, y, fs, hset, kwargs_psd):
    """
    Function to compute irasa for coherence data - Cohrasa
    """

    # TODO: add safety checks

    # Calculate the original coherence over the whole data
    freq, cxy = dsp.coherence(x, y, fs=fs, **kwargs_psd)

    # Start the IRASA procedure
    cxys = np.zeros((len(hset), *cxy.shape))

    for i, h in enumerate(hset):
        # Get the upsampling/downsampling (h, 1/h) factors as integer
        rat = fractions.Fraction(str(h))
        up, down = rat.numerator, rat.denominator
        # resample data
        data_up_x = dsp.resample_poly(x, up, down, axis=-1)
        data_down_x = dsp.resample_poly(x, down, up, axis=-1)

        data_up_y = dsp.resample_poly(y, up, down, axis=-1)
        data_down_y = dsp.resample_poly(y, down, up, axis=-1)

        # Calculate the coherence using same params as original
        _, coh_up = dsp.coherence(data_up_x, data_up_y, h * fs, **kwargs_psd)
        _, coh_down = dsp.coherence(data_down_x, data_down_y, fs / h, **kwargs_psd)
        # Geometric mean of h and 1/h
        cxys[i, :] = np.sqrt(coh_up * coh_down)

    # median resampled data
    Cxy_aperiodic = np.median(cxys, axis=0)
    Cxy_periodic = np.abs(cxy - Cxy_aperiodic)

    return Cxy_periodic, Cxy_aperiodic, freq


# %% temporal cohrasa



"""Custom classes for pyrasa."""

from typing import Protocol, TypedDict

import numpy as np
import pandas as pd
from attrs import define


class IrasaFun(Protocol):
    """
    A protocol defining the interface for an IRASA function used in the PyRASA library.

    The `IrasaFun` protocol specifies the expected signature of a function used to separate
    aperiodic and periodic components of a power spectrum using the IRASA algorithm.
    Any function conforming to this protocol can be passed to other PyRASA functions
    as a custom IRASA implementation.

    Methods
    -------
    __call__(data: np.ndarray, fs: int, h: float,
             up_down: str | None, time_orig: np.ndarray | None = None) -> np.ndarray
        Separates the input data into its aperiodic and periodic components based on the IRASA method.

    Parameters
    ----------
    data : np.ndarray
        The input time series data to be analyzed.
    fs : int
        The sampling frequency of the input data.
    h : float
        The resampling factor used in the IRASA algorithm.
    up_down : str | None
        A string indicating the direction of resampling ('up' or 'down').
        If None, no resampling is performed.
    time_orig : np.ndarray | None, optional
        The original time points of the data, used for interpolation if necessary.
        If None, no interpolation is performed.

    Returns
    -------
    np.ndarray
        The output of the IRASA function.
    """

    def __call__(
        self, data: np.ndarray, fs: int, h: float, up_down: str | None, time_orig: np.ndarray | None = None
    ) -> np.ndarray: ...


class IrasaSprintKwargsTyped(TypedDict):
    """TypedDict for the IRASA sprint function."""

    nfft: int
    hop: int
    win_duration: float
    dpss_settings: dict
    win_kwargs: dict


@define
class AperiodicFit:
    """Container for the results of aperiodic model fits."""

    aperiodic_params: pd.DataFrame
    gof: pd.DataFrame
    model: pd.DataFrame



"""Utilities for slope fitting."""

import warnings
from collections.abc import Iterable

import numpy as np
import pandas as pd

from pyrasa.utils.fit_funcs import AbstractFitFun, FixedFitFun, KneeFitFun
from pyrasa.utils.types import AperiodicFit


def _compute_aperiodic_model(
    aperiodic_spectrum: np.ndarray,
    freq: np.ndarray,
    fit_func: str | type[AbstractFitFun],
    scale_factor: float | int = 1,
) -> tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """helper function to model the aperiodic spectrum"""

    if isinstance(fit_func, str):
        if fit_func == 'fixed':
            fit_func = FixedFitFun
        elif fit_func == 'knee':
            fit_func = KneeFitFun
        else:
            raise ValueError('fit_func should be either a string ("fixed", "knee") or of type AbastractFitFun')

    fit_f = fit_func(freq, aperiodic_spectrum, scale_factor=scale_factor)
    params, gof, pred = fit_f.fit_func()

    return params, gof, pred


def compute_aperiodic_model(
    aperiodic_spectrum: np.ndarray,
    freqs: np.ndarray,
    fit_func: str | type[AbstractFitFun] = 'fixed',
    ch_names: Iterable | None = None,
    scale: bool = False,
    fit_bounds: tuple[float, float] | None = None,
) -> AperiodicFit:
    """
    Computes aperiodic parameters from the aperiodic spectrum using scipy's curve fitting function.

    This function can be used to model the aperiodic (1/f-like) component of the power spectrum. Per default, users can
    choose between a fixed or knee model fit or specify their own fit method see examples custom_fit_functions.ipynb
    for an example. The function returns the fitted parameters for each channel along with some
    goodness of fit metrics.

    Parameters
    ----------
    aperiodic_spectrum : np.ndarray
        A 1 or 2D array of power values for the aperiodic spectrum where the shape is
        expected to be either (Samples,) or (Channels, Samples).
    freqs : np.ndarray
        A 1D array of frequency values corresponding to the aperiodic spectrum.
    fit_func : str or type[AbstractFitFun], optional
        The fitting function to use. Can be "fixed" for a linear fit or "knee" for a fit that includes a
        knee (bend) in the spectrum or a class that is inherited from AbstractFitFun. The default is 'fixed'.
    ch_names : Iterable or None, optional
        Channel names corresponding to the aperiodic spectrum. If None, channels will be named numerically
        in ascending order. Default is None.
    scale : bool, optional
        Whether to scale the data to improve fitting accuracy. This is useful in cases where
        power values are very small (e.g., 1e-28), which may lead to numerical precision issues during fitting.
        After fitting, the parameters are rescaled to match the original data scale. Default is False.
    fit_bounds : tuple[float, float] or None, optional
        Tuple specifying the lower and upper frequency bounds for the fit function. If None, the entire frequency
        range is used. Otherwise, the spectrum is cropped to the specified bounds. Default is None.

    Returns
    -------
    AperiodicFit
        An object containing three pandas DataFrames:
            - aperiodic_params : pd.DataFrame
                A DataFrame containing the fitted aperiodic parameters for each channel.
            - gof : pd.DataFrame
                A DataFrame containing the goodness of fit metrics for each channel.
            - model : pd.DataFrame
                A DataFrame containing the predicted aperiodic model for each channel.

    Notes
    -----
    This function fits the aperiodic component of the power spectrum using scipy's curve fitting function.
    The fitting can be performed using either a simple linear model ('fixed') or a more complex model
    that includes a "knee" point, where the spectrum bends. The resulting parameters can help in
    understanding the underlying characteristics of the aperiodic component in the data.

    If the `fit_bounds` parameter is used, it ensures that only the specified frequency range is considered
    for fitting, which can be important to avoid fitting artifacts outside the region of interest.

    The `scale` parameter can be crucial when dealing with data that have extremely small values,
    as it helps to mitigate issues related to machine precision during the fitting process.

    The function asserts that the input data are of the correct type and shape, and raises warnings
    if the first frequency value is zero, as this can cause issues during model fitting.
    """

    assert isinstance(aperiodic_spectrum, np.ndarray), 'aperiodic_spectrum should be a numpy array.'
    if aperiodic_spectrum.ndim == 1:
        aperiodic_spectrum = aperiodic_spectrum[np.newaxis, :]
    assert aperiodic_spectrum.ndim == 2, 'Data shape needs to be either of shape (Channels, Samples) or (Samples, ).'  # noqa PLR2004

    assert isinstance(freqs, np.ndarray), 'freqs should be a numpy array.'
    assert freqs.ndim == 1, 'freqs needs to be of shape (freqs,).'

    assert isinstance(
        ch_names, list | tuple | np.ndarray | None
    ), 'Channel names should be of type list, tuple or numpy.ndarray or None'

    if fit_bounds is not None:
        fmin, fmax = freqs.min(), freqs.max()
        assert fit_bounds[0] > fmin, f'The selected lower bound is lower than the lowest frequency of {fmin}Hz'
        assert fit_bounds[1] < fmax, f'The selected upper bound is higher than the highest frequency of {fmax}Hz'
        freq_logical = np.logical_and(freqs >= fit_bounds[0], freqs <= fit_bounds[1])
        aperiodic_spectrum, freqs = aperiodic_spectrum[:, freq_logical], freqs[freq_logical]

    if freqs[0] == 0:
        warnings.warn(
            'The first frequency appears to be 0 this will result in aperiodic model fitting problems. '
            + 'Frequencies will be evaluated starting from the next highest in Hz'
        )
        freqs = freqs[1:]
        aperiodic_spectrum = aperiodic_spectrum[:, 1:]

    # generate channel names if not given
    if ch_names is None:
        ch_names = [str(i) for i in np.arange(aperiodic_spectrum.shape[0])]

    if scale:

        def num_zeros(decimal: int) -> float:
            return np.inf if decimal == 0 else -np.floor(np.log10(abs(decimal))) - 1

        scale_factor = 10 ** num_zeros(aperiodic_spectrum.min())
        aperiodic_spectrum = aperiodic_spectrum * scale_factor
    else:
        scale_factor = 1

    ap_list, gof_list, pred_list = [], [], []
    for ix, ch_name in enumerate(ch_names):
        params, gof, pred = _compute_aperiodic_model(
            aperiodic_spectrum=aperiodic_spectrum[ix],
            freq=freqs,
            fit_func=fit_func,
            scale_factor=scale_factor,
        )

        params['ch_name'] = ch_name
        gof['ch_name'] = ch_name
        pred['ch_name'] = ch_name

        ap_list.append(params)
        gof_list.append(gof)
        pred_list.append(pred)

    # combine & return
    return AperiodicFit(aperiodic_params=pd.concat(ap_list), gof=pd.concat(gof_list), model=pd.concat(pred_list))


def compute_aperiodic_model_sprint(
    aperiodic_spectrum: np.ndarray,
    freqs: np.ndarray,
    times: np.ndarray,
    fit_func: str | type[AbstractFitFun] = 'fixed',
    scale: bool = False,
    ch_names: Iterable | None = None,
    fit_bounds: tuple[float, float] | None = None,
) -> AperiodicFit:
    """
    Extracts aperiodic parameters from the aperiodic spectrogram using scipy's curve fitting
    function.

    This function computes aperiodic parameters for each time point in the spectrogram by applying either one of
    two different curve fitting functions (`fixed` or `knee`) or a custom function specified by user to the data.
    See examples custom_fit_functions.ipynb. The parameters, along with the goodness of
    fit for each time point, are returned in a concatenated format.

    Parameters
    ----------
    aperiodic_spectrum : np.ndarray
        A 2 or 3D array of power values from the aperiodic spectrogram, with shape (Frequencies, Time)
        or (Channels, Frequencies, Time).
    freqs : np.ndarray
        A 1D array of frequency values corresponding to the aperiodic spectrogram.
    times : np.ndarray
        A 1D array of time values corresponding to the aperiodic spectrogram.
    fit_func : str or type[AbstractFitFun], optional
        The fitting function to use. Can be "fixed" for a linear fit or "knee" for a fit that includes a
        knee (bend) in the spectrum or a class that is inherited from AbstractFitFun. The default is 'fixed'..
    ch_names : Iterable or None, optional
        Channel names corresponding to the aperiodic spectrogram. If None, channels will be named numerically
        in ascending order. Default is None.
    scale : bool, optional
        Whether to scale the data to improve fitting accuracy. This is useful when fitting a knee in cases where
        power values are very small, leading to numerical precision issues. Default is False.
    fit_bounds : tuple[float, float] or None, optional
        Tuple specifying the lower and upper frequency bounds for the fit function. If None, the entire frequency
        range is used. Otherwise, the spectrum is cropped to the specified bounds before fitting. Default is None.

    Returns
    -------
    AperiodicFit
        An object containing three pandas DataFrames:
            - aperiodic_params : pd.DataFrame
                A DataFrame containing the aperiodic parameters (e.g., Offset and Exponent)
                for each channel and each time point.
            - gof : pd.DataFrame
                A DataFrame containing the goodness of fit metrics for each channel and each time point.
            - model : pd.DataFrame
                A DataFrame containing the predicted aperiodic model for each channel and each time point.

    Notes
    -----
    This function iterates over each time point in the provided spectrogram to extract aperiodic parameters
    using the specified fit function. It leverages the `compute_aperiodic_model` function for individual fits
    at each time point, then combines the results across all time points into comprehensive DataFrames.

    The `fit_bounds` parameter allows for frequency range restrictions during fitting, which can help in focusing
    the analysis on a particular frequency band of interest.

    Scaling the data using the `scale` parameter can be particularly important when dealing with very small power
    values that might lead to poor fitting due to numerical precision limitations.

    """

    ap_t_list, gof_t_list, pred_t_list = [], [], []

    for ix, t in enumerate(times):
        aperiodic_fit = compute_aperiodic_model(
            aperiodic_spectrum[:, :, ix],
            freqs=freqs,
            fit_func=fit_func,
            ch_names=ch_names,
            fit_bounds=fit_bounds,
            scale=scale,
        )
        aperiodic_fit.aperiodic_params['time'] = t
        aperiodic_fit.gof['time'] = t
        aperiodic_fit.model['time'] = t

        ap_t_list.append(aperiodic_fit.aperiodic_params)
        gof_t_list.append(aperiodic_fit.gof)
        pred_t_list.append(aperiodic_fit.model)

    return AperiodicFit(aperiodic_params=pd.concat(ap_t_list), gof=pd.concat(gof_t_list), model=pd.concat(pred_t_list))



# %%
from neurodsp.sim import sim_combined
from neurodsp.utils import create_times
import scipy.signal as dsp
import matplotlib.pyplot as plt
import seaborn as sns
from pyrasa import irasa

sns.set_style('ticks')
sns.set_context('poster')
# %%
fs = 500
n_seconds = 60
duration = 2
overlap = 0.5

sim_components = {'sim_powerlaw': {'exponent': -1}, 'sim_oscillation': {'freq': 10}}
# %%

sig = sim_combined(n_seconds=n_seconds, fs=fs, components=sim_components)
times = create_times(n_seconds=n_seconds, fs=fs)
# %%
max_times = times < 1
f, axes = plt.subplots(ncols=2, figsize=(8, 4))
axes[0].plot(times[max_times], sig[max_times])
axes[0].set_ylabel('Amplitude (a.u.)')
axes[0].set_xlabel('Time (s)')
freq, psd = dsp.welch(sig, fs=fs, nperseg=duration * fs, noverlap=duration * fs * overlap)
axes[1].loglog(freq, psd)
axes[1].set_ylabel('Power (a.u.)')
axes[1].set_xlabel('Frequency (Hz)')

plt.tight_layout()


from pyrasa.irasa import irasa, irasa_sprint

# %%
irasa_psd = irasa(
    sig,
    fs=fs,
    band=(1, 50),
    psd_kwargs={'nperseg': duration * fs, 'noverlap': duration * fs * overlap},
    hset_info=(1, 4, 0.05),
)

#%%
duration = 5

irasa_out_tf = irasa_sprint(
    sig,
    fs=fs,
    band=(1, 50),
    win_duration=duration,
    overlap_fraction=.9,
    hset_info=(1, 2, 0.1),
)
#%%
from neurodsp.plts import plot_timefrequency
plot_timefrequency(times=irasa_out_tf.time, 
                   freqs=irasa_out_tf.freqs, 
                   powers=irasa_out_tf.periodic)

#%%
plot_timefrequency(times=irasa_out_tf.time, 
                   freqs=irasa_out_tf.freqs, 
                   powers=irasa_out_tf.aperiodic)


# %%
f, axes = plt.subplots(ncols=2, figsize=(8, 4))
axes[0].set_title('Periodic')
axes[0].plot(freq_irasa, psd_p[0, :])
axes[0].set_ylabel('Power (a.u.)')
axes[0].set_xlabel('Frequency (Hz)')
axes[1].set_title('Aperiodic')
axes[1].loglog(freq_irasa, psd_ap[0, :])
axes[1].set_ylabel('Power (a.u.)')
axes[1].set_xlabel('Frequency (Hz)')

f.tight_layout()
# %% get periodic stuff
from pyrasa.utils.peak_utils import get_peak_params

get_peak_params(psd_p, freqs=freq_irasa)
# %% get aperiodic stuff
from pyrasa.utils.aperiodic_utils import compute_slope

ap_params, gof_params = compute_slope(aperiodic_spectrum=psd_ap, freqs=freq_irasa, fit_func='knee')
ap_params
# %%
gof_params
# %% Lets check the knee
knee_freq = 15
knee = knee_freq ** (2 * 0.5 + 1.5)

sim_components = {'sim_knee': {'exponent1': -0.5, 'exponent2': -1.5, 'knee': knee}, 'sim_oscillation': {'freq': 10}}

sig = sim_combined(n_seconds=n_seconds, fs=fs, components=sim_components)
# %%
max_times = times < 1
f, axes = plt.subplots(ncols=2, figsize=(8, 4))
axes[0].plot(times[max_times], sig[max_times])
axes[0].set_ylabel('Amplitude (a.u.)')
axes[0].set_xlabel('Time (s)')
freq, psd = dsp.welch(sig, fs=fs, nperseg=duration * fs, noverlap=duration * fs * overlap)
axes[1].loglog(freq, psd)
axes[1].set_ylabel('Power (a.u.)')
axes[1].set_xlabel('Frequency (Hz)')

plt.tight_layout()
# %%
freq_irasa, psd_ap, psd_p = irasa(
    sig,
    fs=fs,
    band=(1, 150),
    kwargs_psd={'nperseg': duration * fs, 'noverlap': duration * fs * overlap},
    hset_info=(1, 2, 0.025),
)
# %%
f, axes = plt.subplots(ncols=2, figsize=(8, 4))
axes[0].set_title('Periodic')
axes[0].plot(freq_irasa, psd_p[0, :])
axes[0].set_ylabel('Power (a.u.)')
axes[0].set_xlabel('Frequency (Hz)')
axes[1].set_title('Aperiodic')
axes[1].loglog(freq_irasa, psd_ap[0, :])
axes[1].set_ylabel('Power (a.u.)')
axes[1].set_xlabel('Frequency (Hz)')

f.tight_layout()

# %% get periodic stuff
from pyrasa.utils.peak_utils import get_peak_params

get_peak_params(psd_p, freqs=freq_irasa)
# %% get aperiodic stuff
from pyrasa.utils.aperiodic_utils import compute_slope

ap_params, gof_params = compute_slope(aperiodic_spectrum=psd_ap, freqs=freq_irasa, fit_func='fixed')
ap_params
# %%
gof_params
# %%
ap_params, gof_params = compute_slope(aperiodic_spectrum=psd_ap, freqs=freq_irasa, fit_func='knee')
ap_params
# %%
gof_params
# %%



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information


from sphinx.config import is_serializable

import pyrasa

# import pyrasa.irasa_mne

project = 'PyRASA'
copyright = '2024, Fabian Schmidt, Thomas Hartmann'
author = 'Fabian Schmidt, Thomas Hartmann'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration


extensions = [
    'sphinx.ext.githubpages',
    'sphinx.ext.autodoc',
    'sphinx.ext.viewcode',
    'sphinx.ext.autosummary',
    'sphinx.ext.intersphinx',
    'sphinx.ext.napoleon',
    'numpydoc',
    'sphinx_gallery.gen_gallery',
    'sphinx.ext.mathjax',  # optional, if you need to render math
    'sphinx.ext.viewcode',
    'nbsphinx',
]


numpydoc_xref_param_type = True
numpydoc_class_members_toctree = False
numpydoc_attributes_as_param_list = True
numpydoc_xref_aliases = {
    'array-like': ':term:`array_like <numpy:array_like>`',
    'int': ':class:`int <python:int>`',
    'bool': ':class:`bool <python:bool>`',
    'float': ':class:`float <python:float>`',
    'list': ':class:`list <python:list>`',
    'tuple': ':class:`tuple <python:tuple>`',
}
numpydoc_xref_ignore = {
    # words
    'instance',
    'instances',
    'of',
}


# generate autosummary even if no references
autosummary_generate = True
autodoc_default_options = {'inherited-members': None}
default_role = 'autolink'  # XXX silently allows bad syntax, someone should fix

exclude_patterns = ['auto_examples/index.rst', '_build', 'Thumbs.db', '.DS_Store', 'generated']

html_show_sourcelink = False
html_copy_source = False

html_theme = 'pydata_sphinx_theme'


templates_path = ['_templates']
html_static_path = ['_static']
# html_css_files = ["style.css"]

source_suffix = ['.rst', '.md']
# master_doc = "index"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = 'dev' if 'dev' in pyrasa.__version__ else pyrasa.__version__
# The full version, including alpha/beta/rc tags.
release = version

html_title = 'PyRASA' + ' ' + release
html_short_title = 'PyRASA'

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output


html_theme_options = {
    #'navbar_sidebarrel': False,
    'icon_links': [
        dict(
            name='GitHub',
            url='https://github.com/schmidtfa/pyrasa',
            icon='fab fa-github-square',
        ),
    ],
    'icon_links_label': 'Quick Links',  # for screen reader
    'use_edit_page_button': False,
    'navigation_with_keys': False,
    'show_toc_level': 1,
    'header_links_before_dropdown': 6,
    'navbar_end': ['theme-switcher', 'navbar-icon-links'],
}

html_context = {
    'default_mode': 'auto',
    'doc_path': 'doc',
}

html_sidebars = {}


html_short_title = 'PyRASA'


sphinx_gallery_conf = {
    'doc_module': 'pyrasa',
    'reference_url': {
        'pyrasa': None,
    },
    'backreferences_dir': 'generated',
    'examples_dirs': 'examples',
    'within_subsection_order': 'ExampleTitleSortKey',
    'gallery_dirs': 'auto_examples',
    'filename_pattern': '^((?!sgskip).)*$',
}

assert is_serializable(sphinx_gallery_conf)



# %%
import numpy as np
from cluster_jobs.simple_simulation import simple_simulation
from plus_slurm import JobCluster, PermuteArgument

# % get jobcluster
job_kwargs = {
    'required_ram': '2G',
    'request_cpus': 1,
    'request_time': 30,
    'python_bin': '/home/schmidtfa/miniconda3/envs/pyrasa/bin/python',
}

# % put in jobs...
# aperiodic parameters
EXPS = np.round(np.arange(-3, 0.5, 0.5), 1).tolist()
EXPS_KNEE = np.round(np.arange(-4, -1, 0.2), 1).tolist()  # need to start above 1
KNEE_FREQS = np.arange(5, 40, 5).tolist()

# Freq related parameters
FREQS = np.arange(5, 36, 1).tolist()
POWERS = np.round(np.arange(0, 2.0, 0.2), 1).tolist()
BWS = np.round(np.arange(0.5, 3.5, 0.5), 1).tolist()

# Burst related parameters
BPROBS = np.round(np.arange(0.2, 0.8, 0.1), 1).tolist()

# %% exponent_delta
job_cluster = JobCluster(**job_kwargs)
job_cluster.add_job(
    simple_simulation,
    subject_id='exponent_delta',  # PermuteArgument(sim_kinds),
    exp=PermuteArgument(EXPS),
    fit_type='fixed',
)
job_cluster.submit(do_submit=True)

# %% knee_delta
job_cluster = JobCluster(**job_kwargs)
job_cluster.add_job(
    simple_simulation,
    subject_id='knee_delta',  # PermuteArgument(sim_kinds),
    knee_freq=PermuteArgument(KNEE_FREQS),
    fit_type='knee',
)
job_cluster.submit(do_submit=True)

# %% exponent_knee_delta
job_cluster = JobCluster(**job_kwargs)
job_cluster.add_job(
    simple_simulation,
    subject_id='exponent_knee_delta',  # PermuteArgument(sim_kinds),
    exp=PermuteArgument(EXPS_KNEE),
    fit_type='knee',
)
job_cluster.submit(do_submit=True)

# %%burst_delta
job_cluster = JobCluster(**job_kwargs)
job_cluster.add_job(
    simple_simulation,
    subject_id='burst_delta',  # PermuteArgument(sim_kinds),
    b_proba=PermuteArgument(BPROBS),
    fit_type='fixed',
)
job_cluster.submit(do_submit=True)

# %% freq_width_delta
job_cluster = JobCluster(**job_kwargs)
job_cluster.add_job(
    simple_simulation,
    subject_id='freq_width_delta',  # PermuteArgument(sim_kinds),
    bw=PermuteArgument(BWS),
    fit_type='fixed',
)
job_cluster.submit(do_submit=True)

# %% freq_height_delta
job_cluster = JobCluster(**job_kwargs)
job_cluster.add_job(
    simple_simulation,
    subject_id='freq_height_delta',  # PermuteArgument(sim_kinds),
    height=PermuteArgument(POWERS),
    fit_type='fixed',
)
job_cluster.submit(do_submit=True)

# %% freq_delta
job_cluster = JobCluster(**job_kwargs)
job_cluster.add_job(
    simple_simulation,
    subject_id='freq_delta',  # PermuteArgument(sim_kinds),
    freq=PermuteArgument(FREQS),
    fit_type='fixed',
)
job_cluster.submit(do_submit=True)

# %%



"""Utilities for signal decompositon using IRASA"""

import fractions
from collections.abc import Callable
from copy import copy

import numpy as np
import scipy.signal as dsp

from pyrasa.utils.types import IrasaFun


def _gen_irasa(
    data: np.ndarray,
    orig_spectrum: np.ndarray,
    fs: int,
    irasa_fun: IrasaFun,
    hset: np.ndarray,
    time: np.ndarray | None = None,
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Generate original, aperiodic, and periodic spectra using the IRASA algorithm.

    This function implements the IRASA (Irregular Resampling Auto-Spectral Analysis) algorithm
    to decompose a power or cross-spectral density into its periodic and aperiodic components.

    Parameters
    ----------
    data : np.ndarray
        The input time-series data, typically with shape (n_channels, n_times) or similar.
    orig_spectrum : np.ndarray
        The original power spectral density from which periodic and aperiodic components are to be extracted.
    fs : int
        The sampling frequency of the input data in Hz.
    irasa_fun : IrasaFun
        A custom function used to compute power spectral densities. This function should
        take resampled data and return the corresponding spectrum.
    hset : np.ndarray
        An array of up/downsampling factors (e.g., [1.1, 1.2, 1.3, ...]) used in the IRASA algorithm.
    time : np.ndarray | None, optional
        The time vector associated with the original data. This is only necessary if the IRASA function
        requires the time stamps of the original data.

    Returns
    -------
    tuple[np.ndarray, np.ndarray, np.ndarray]
        A tuple containing:
        - `orig_spectrum` (np.ndarray): The original spectrum provided as input.
        - `aperiodic_spectrum` (np.ndarray): The median of the geometric mean of up/downsampled spectra,
        representing the aperiodic component.
        - `periodic_spectrum` (np.ndarray): The difference between the original and the aperiodic spectrum,
        representing the periodic component.

    Notes
    -----
    This implementation of the IRASA algorithm is based on the `yasa.irasa` function from (Vallat & Walker, 2021).
    The IRASA algorithm involves upsampling and downsampling the time-series data by a set of factors (`hset`),
    calculating the power spectra of these resampled data, and then taking the geometric mean of the upsampled
    and downsampled spectra to isolate the aperiodic component.

    References
    ----------
    [1] Vallat, Raphael, and Matthew P. Walker. “An open-source, high-performance tool for automated sleep staging.”
        Elife 10 (2021). doi: https://doi.org/10.7554/eLife.70092
    """

    spectra = np.zeros((len(hset), *orig_spectrum.shape))
    for i, h in enumerate(hset):
        rat = fractions.Fraction(str(h))
        up, down = rat.numerator, rat.denominator

        # Much faster than FFT-based resampling
        data_up = dsp.resample_poly(data, up, down, axis=-1)
        data_down = dsp.resample_poly(data, down, up, axis=-1)

        # Calculate an up/downsampled version of the PSD using same params as original
        spectrum_up = irasa_fun(data=data_up, fs=int(fs * h), h=h, time_orig=time, up_down='up')
        spectrum_dw = irasa_fun(data=data_down, fs=int(fs / h), h=h, time_orig=time, up_down='down')

        # geometric mean between up and downsampled
        spectra[i, :, :] = np.sqrt(spectrum_up * spectrum_dw)

    aperiodic_spectrum = np.median(spectra, axis=0)
    periodic_spectrum = orig_spectrum - aperiodic_spectrum
    return orig_spectrum, aperiodic_spectrum, periodic_spectrum


def _crop_data(
    band: list | tuple,
    freqs: np.ndarray,
    psd_aperiodic: np.ndarray,
    psd_periodic: np.ndarray,
    psd: np.ndarray,
    axis: int,
) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Utility function to crop spectra to a defined frequency range"""

    mask_freqs = np.ma.masked_outside(freqs, *band).mask
    freqs = freqs[~mask_freqs]
    psd_aperiodic = np.compress(~mask_freqs, psd_aperiodic, axis=axis)
    psd_periodic = np.compress(~mask_freqs, psd_periodic, axis=axis)
    psd = np.compress(~mask_freqs, psd, axis=axis)

    return freqs, psd_aperiodic, psd_periodic, psd


def _get_windows(
    nperseg: int, dpss_settings: dict, win_func: Callable, win_func_kwargs: dict
) -> tuple[np.ndarray, np.ndarray]:
    """Generate a window function used for tapering"""

    low_bias_ratio = 0.9
    min_time_bandwidth = 2.0
    win_func_kwargs = copy(win_func_kwargs)

    # special settings in case multitapering is required
    if win_func == dsp.windows.dpss:
        time_bandwidth = dpss_settings['time_bandwidth']
        if time_bandwidth > min_time_bandwidth:
            raise ValueError(f'time_bandwidth should be >= {min_time_bandwidth} for good tapers')

        n_taps = int(np.floor(time_bandwidth - 1))
        win_func_kwargs.update(
            {
                'NW': time_bandwidth / 2,  # half width
                'Kmax': n_taps,
                'sym': False,
                'return_ratios': True,
            }
        )
        win, ratios = win_func(nperseg, **win_func_kwargs)
        if dpss_settings['low_bias']:
            win = win[ratios > low_bias_ratio]
            ratios = ratios[ratios > low_bias_ratio]
    else:
        win = [win_func(nperseg, **win_func_kwargs)]
        ratios = None

    return win, ratios


def _check_irasa_settings(irasa_params: dict, hset_info: tuple) -> None:
    """Check if the input parameters for irasa are specified correctly"""

    valid_hset_shape = 3
    assert isinstance(irasa_params['data'], np.ndarray), 'Data should be a numpy array.'

    # check if hset is specified correctly
    assert isinstance(
        hset_info, tuple | list | np.ndarray
    ), 'hset should be a tuple, list or numpy array of (min, max, step)'
    assert np.shape(hset_info)[0] == valid_hset_shape, 'shape of hset_info should be 3 i.e. (min, max, step)'

    # check that evaluated range fits with the data settings
    nyquist = irasa_params['fs'] / 2
    hmax = np.max(hset_info)
    band_evaluated: tuple[float, float] = (irasa_params['band'][0] / hmax, irasa_params['band'][1] * hmax)
    assert band_evaluated[0] > 0, 'The evaluated frequency range is 0 or lower this makes no sense'
    assert band_evaluated[1] < nyquist, (
        f'The evaluated frequency range goes up to {np.round(band_evaluated[1], 2)}Hz '
        f'which is higher than the Nyquist frequency for your data of {nyquist}Hz, \n'
        'try to either lower the upper bound for the hset or decrease the upper band limit, when running IRASA.'
    )

    filter_settings: list[float] = list(irasa_params['filter_settings'])
    if filter_settings[0] is None:
        filter_settings[0] = band_evaluated[0]
    if filter_settings[1] is None:
        filter_settings[1] = band_evaluated[1]

    assert np.logical_and(band_evaluated[0] >= filter_settings[0], band_evaluated[1] <= filter_settings[1]), (
        f'You run IRASA in a frequency range from '
        f'{np.round(band_evaluated[0], irasa_params["hset_accuracy"])} - '
        f'{np.round(band_evaluated[1], irasa_params["hset_accuracy"])}Hz. \n'
        'Your settings specified in "filter_settings" indicate that you have a pass band from '
        f'{np.round(filter_settings[0], irasa_params["hset_accuracy"])} - '
        f'{np.round(filter_settings[1], irasa_params["hset_accuracy"])}Hz. \n'
        'This means that your evaluated range likely contains filter artifacts. \n'
        'Either change your filter settings, adjust hset or the parameter "band" accordingly. \n'
        f'You want to make sure that the lower band limit divided by the upper bound of the hset '
        f'> {np.round(filter_settings[0], irasa_params["hset_accuracy"])} \n'
        'and that upper band limit times the upper bound of the hset < '
        f'{np.round(filter_settings[1], irasa_params["hset_accuracy"])}'
    )


# Calculate original spectrum
def _compute_psd_welch(
    data: np.ndarray,
    fs: int,
    nperseg: int | None,
    win_kwargs: dict,
    dpss_settings: dict,
    noverlap: int | None = None,
    nfft: int | None = None,
    detrend: str = 'constant',
    return_onesided: bool = True,
    scaling: str = 'density',
    axis: int = -1,
    average: str = 'mean',
) -> tuple[np.ndarray, np.ndarray]:
    """Compute power spectral densities via scipy.signal.welch"""

    if nperseg is None:
        nperseg = data.shape[-1]
    win, ratios = _get_windows(nperseg, dpss_settings, **win_kwargs)

    psds = []
    for cur_win in win:
        freq, psd = dsp.welch(
            data,
            fs=fs,
            window=cur_win,
            nperseg=nperseg,
            noverlap=noverlap,
            nfft=nfft,
            detrend=detrend,
            return_onesided=return_onesided,
            scaling=scaling,
            axis=axis,
            average=average,
        )
        psds.append(psd)

    if ratios is None:
        psd = np.mean(psds, axis=0)
    else:
        weighted_psds = [ratios[ix] * cur_sgramm for ix, cur_sgramm in enumerate(psds)]
        psd = np.sum(weighted_psds, axis=0) / np.sum(ratios)

    return freq, psd


def _compute_sgramm(  # noqa C901
    x: np.ndarray,
    fs: int,
    nfft: int,
    hop: int,
    win_duration: float,
    dpss_settings: dict,
    win_kwargs: dict,
    h: float = 1.0,
    up_down: str | None = None,
    time_orig: np.ndarray | None = None,
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Compute spectrograms via scipy.signal.stft"""

    nperseg = int(np.floor(fs * win_duration))

    if up_down == 'up':
        hop = int(hop * h)

    if up_down == 'down':
        hop = int(hop / h)

    win, ratios = _get_windows(nperseg, dpss_settings, **win_kwargs)

    sgramms = []
    for cur_win in win:
        freq, time, sgramm = dsp.stft(
            x, nfft=nfft, nperseg=nperseg, noverlap=nperseg - hop, fs=fs, window=cur_win, scaling='psd'
        )
        sgramm = np.abs(sgramm) ** 2
        sgramms.append(sgramm)

    if ratios is None:
        sgramm = np.mean(sgramms, axis=0)
    else:
        weighted_sgramms = [ratios[ix] * cur_sgramm for ix, cur_sgramm in enumerate(sgramms)]
        sgramm = np.sum(weighted_sgramms, axis=0) / np.sum(ratios)

    if time_orig is not None:
        sgramm = sgramm[:, :, : time_orig.shape[-1]]

    sgramm = np.squeeze(sgramm)

    return freq, time, sgramm



#%%
import sys
from neurodsp.sim import set_random_seed
from neurodsp.sim import sim_powerlaw, sim_oscillation
from neurodsp.utils import create_times
from neurodsp.plts import plot_timefrequency#

from neurodsp.timefrequency import compute_wavelet_transform
import numpy as np
import matplotlib.pyplot as plt
#import seaborn as sns
import pandas as pd

import matplotlib as mpl
new_rc_params = {'text.usetex': False,
                 "svg.fonttype": 'none'
}
mpl.rcParams.update(new_rc_params)

set_random_seed(84)

from pyrasa.irasa import irasa_sprint
# %%
# Set some general settings, to be used across all simulations
fs = 500
n_seconds = 15
duration=4
overlap=0.5

# Create a times vector for the simulations
times = create_times(n_seconds, fs)


alpha = sim_oscillation(n_seconds=.5, fs=fs, freq=10)
no_alpha = np.zeros(len(alpha))
beta = sim_oscillation(n_seconds=.5, fs=fs, freq=25)
no_beta = np.zeros(len(beta))

exp_1 = sim_powerlaw(n_seconds=2.5, fs=fs, exponent=-1)
exp_2 = sim_powerlaw(n_seconds=2.5, fs=fs, exponent=-2)


alphas = np.concatenate([no_alpha, alpha, no_alpha, alpha, no_alpha])
betas = np.concatenate([beta, no_beta, beta, no_beta, beta])

sim_ts = np.concatenate([exp_1 + alphas, 
                         exp_1 + alphas + betas, 
                         exp_1 + betas, 
                         exp_2 + alphas, 
                         exp_2 + alphas + betas, 
                         exp_2 + betas, ])

# %%
freqs = np.arange(1, 50, 0.5)
import scipy.signal as dsp

irasa_sprint_spectrum = irasa_sprint(sim_ts,#np.array([sim_ts, sim_ts]), 
                                    fs=fs,
                                    band=(1, 50),
                                    overlap_fraction=.95,
                                    win_duration=.5,
                                    ch_names=['A'],
                                    hset_info=(1.05, 4., 0.05),
                                    win_func=dsp.windows.hann)
# %%
peak_kwargs = { 'smooth': True,
                'smoothing_window':1,
                'peak_threshold':5,
                'min_peak_height':.01,
                'peak_width_limits': (0.5, 12)}
ap_error = irasa_sprint_spectrum.get_aperiodic_error(peak_kwargs)
# %%
plt.plot(ap_error[0,:,:].mean(axis=1))
# %%



#%%
from neurodsp.sim import sim_combined
from neurodsp.utils import create_times
import numpy as np
import scipy.signal as dsp


fs = 1000
n_seconds = 60
duration=2
overlap=0.5

sim_components = {'sim_powerlaw': {'exponent' : -1}, 
                  'sim_oscillation': {'freq' : 10}}


sig = sim_combined(n_seconds=n_seconds, fs=fs, components=sim_components)
times = create_times(n_seconds=n_seconds, fs=fs)
#%%
import fractions
kwargs_psd = {'nperseg': duration*fs, 
              'noverlap': duration*fs*overlap}

resampling_factor = 1.5

rat = fractions.Fraction(str(resampling_factor))
up, down = rat.numerator, rat.denominator

# Much faster than FFT-based resampling
data_up = dsp.resample_poly(sig, up, down, axis=-1)
data_down = dsp.resample_poly(sig, down, up, axis=-1)

# Calculate an up/downsampled version of the PSD using same params as original
win_duration = 2
hop = 100

nperseg = int(np.floor(fs * win_duration))

win = dsp.windows.hann(nperseg)

SFT = dsp.ShortTimeFFT(win, hop=hop, fs=fs, scale_to='psd')
t_inc = SFT.T
psd = SFT.spectrogram(sig, detr='constant')

hop_up = int(hop * resampling_factor)
SFT_u = dsp.ShortTimeFFT.from_window('hann', 
                                     nperseg=nperseg, 
                                     fs=fs * resampling_factor, 
                                     noverlap=nperseg-hop_up)
psd_up = SFT_u.stft(data_up)


#%%
hop_dw = int(hop / resampling_factor)
noverlap=nperseg-hop_dw
N = len(data_down)
SFT_d = dsp.ShortTimeFFT.from_window('hann', 
                                     nperseg=nperseg, 
                                     fs=fs / resampling_factor, 
                                     noverlap=nperseg-hop_dw,
                                     fft_mode='centered',)
psd_dw = SFT_d.spectrogram(data_down, p0=0, p1=(N-noverlap)//SFT_d.hop, k_offset=N//2)

# %%
psd_dw.shape

# %%
psd.shape
# %%
psd_up.shape


# %%
import numpy as np
import scipy.signal as dsp
import fractions
from neurodsp.sim import sim_oscillation, sim_powerlaw

# Example signal and parameters
# Set some general settings, to be used across all simulations
fs = 500
n_seconds = 15
duration=4
overlap=0.5

# Create a times vector for the simulations
#times = create_times(n_seconds, fs)


alpha = sim_oscillation(n_seconds=.5, fs=fs, freq=10)
no_alpha = np.zeros(len(alpha))
beta = sim_oscillation(n_seconds=.5, fs=fs, freq=25)
no_beta = np.zeros(len(beta))

exp_1 = sim_powerlaw(n_seconds=2.5, fs=fs, exponent=-1)
exp_2 = sim_powerlaw(n_seconds=2.5, fs=fs, exponent=-2)


alphas = np.concatenate([no_alpha, alpha, no_alpha, alpha, no_alpha])
betas = np.concatenate([beta, no_beta, beta, no_beta, beta])

sig = np.concatenate([exp_1 + alphas, 
                         exp_1 + alphas + betas, 
                         exp_1 + betas, 
                         exp_2 + alphas, 
                         exp_2 + alphas + betas, 
                         exp_2 + betas, ])

resampling_factor = 1.1  # Resampling factor
win_duration = 1  # Window duration for spectrogram
hop = 10  # Hop size for spectrogram
hset = np.arange(1, 2, 0.05).round(2)
freq_res = 0.5
nfft = int(fs / freq_res)

# Original spectrogram
# Window and hop size for spectrogram
nperseg = int(np.floor(fs * win_duration))
win = dsp.windows.hann(nperseg)
freq, t, psd = dsp.stft(sig, fs=fs, nfft=nfft, window=win, nperseg=nperseg, 
            noverlap=nperseg - hop, scaling='psd')

psd = (np.abs(psd) ** 2)

average_psd = np.zeros([len(hset), *psd.shape])

for i, resampling_factor in enumerate(hset):
    # Calculate resampling factors
    rat = fractions.Fraction(str(resampling_factor))
    up, down = rat.numerator, rat.denominator

    # Resample the signal
    data_up = dsp.resample_poly(sig, up, down, axis=-1)
    data_down = dsp.resample_poly(sig, down, up, axis=-1)

    # Upsampled spectrogram
    hop_up = int(hop * resampling_factor)
    f_up, t_up, psd_up = dsp.stft(data_up, nfft=nfft, fs=fs * resampling_factor, 
                    window=win, nperseg=nperseg, 
                    noverlap=nperseg - hop_up, scaling='psd')

    # Downsampled spectrogram
    hop_down = int(hop / resampling_factor)
    f_dw, t_dw, psd_dw = dsp.stft(data_down, nfft=nfft, fs=fs / resampling_factor, 
                    window=win, nperseg=nperseg, 
                    noverlap=nperseg - hop_down, scaling='psd')

    # Ensure the time axis has the same number of values by adjusting the hop size
    psd_up = psd_up[:, :psd.shape[1]]
    psd_dw = psd_dw[:, :psd.shape[1]]

    # Average the PSDs
    average_psd[i,:,:] = np.sqrt((np.abs(psd_up) ** 2) * (np.abs(psd_dw) ** 2))
    # If needed, the time axis can be derived from the spectrogram output
    #time_axis = np.linspace(0, len(sig) / fs, psd.shape[1])

    print("Average PSD shape:", average_psd.shape)
    #print("Time axis length:", len(time_axis))

aperiodic = np.median(average_psd, axis=0)
periodic = psd - aperiodic

#%%
from neurodsp.plts import plot_timefrequency#
import matplotlib.pyplot as plt
f, axes = plt.subplots(figsize=(14, 4), ncols=3)

fmask = freq < 50

plot_timefrequency(t, freq[fmask], psd[fmask,:], vmin=0, ax=axes[0])
plot_timefrequency(t, freq[fmask], aperiodic[fmask,:], vmin=0, ax=axes[1])
plot_timefrequency(t, freq[fmask], periodic[fmask,:], vmin=0, ax=axes[2])

# %%
from pyrasa.irasa import irasa_sprint

irasa_sprint_spectrum = irasa_sprint(sig[np.newaxis, :], fs=fs,
                                                       band=(1, 100),
                                                       freq_res=.5,
                                                       hop=100,
                                                       win_duration=1.,
                                                       hset_info=(1.05, 2., 0.05))
# %%



"""Functions to compute IRASA."""

from collections.abc import Callable
from typing import TYPE_CHECKING, Any

import numpy as np
import scipy.signal as dsp

from pyrasa.utils.irasa_spectrum import IrasaSpectrum
from pyrasa.utils.irasa_tf_spectrum import IrasaTfSpectrum

# from scipy.stats.mstats import gmean
from pyrasa.utils.irasa_utils import (
    _check_irasa_settings,
    _compute_psd_welch,
    _compute_sgramm,
    _crop_data,
    _gen_irasa,
)

if TYPE_CHECKING:
    from pyrasa.utils.types import IrasaSprintKwargsTyped


# %% irasa
def irasa(
    data: np.ndarray,
    fs: int,
    band: tuple[float, float],
    psd_kwargs: dict,
    ch_names: np.ndarray | None = None,
    win_func: Callable = dsp.windows.hann,
    win_func_kwargs: dict | None = None,
    dpss_settings_time_bandwidth: float = 2.0,
    dpss_settings_low_bias: bool = True,
    dpss_eigenvalue_weighting: bool = True,
    filter_settings: tuple[float | None, float | None] = (None, None),
    hset_info: tuple[float, float, float] = (1.05, 2.0, 0.05),
    hset_accuracy: int = 4,
) -> IrasaSpectrum:
    """
    Computes the aperiodic and periodic components of the power spectrum from a time series using the
    Irregular Resampling Autocorrelation (IRASA) algorithm.

    The IRASA algorithm allows for the decomposition of neural signals into fractal (aperiodic) and
    oscillatory (periodic) components, providing insight into the underlying dynamics of the data.

    Parameters
    ----------
    data : np.ndarray
        Time series data, where the shape is expected to be either (Samples,) or (Channels, Samples).
    fs : int
        Sampling frequency of the data in Hz.
    band : tuple[float, float]
        The frequency range (lower and upper bounds in Hz) over which to compute the spectra.
    psd_kwargs : dict
        Keyword arguments to be passed to the `scipy.signal.welch` function for PSD estimation.
    ch_names : np.ndarray | None, optional
        Channel names associated with the data, if available. Default is None.
    win_func : Callable, optional
        Window function to be used in Welch's method. Default is `dsp.windows.hann`.
    win_func_kwargs : dict | None, optional
        Additional keyword arguments for the window function. Default is None.
    dpss_settings_time_bandwidth : float, optional
        Time-bandwidth product for the DPSS windows if used. Default is 2.0.
    dpss_settings_low_bias : bool, optional
        Keep only tapers with eigenvalues > 0.9. Default is True.
    dpss_eigenvalue_weighting : bool, optional
        Whether or not to apply eigenvalue weighting in DPSS. If True, spectral estimates weighted by
        the concentration ratio of their respective tapers before combining. Default is True.
    filter_settings : tuple[float | None, float | None], optional
        Cutoff frequencies for highpass and lowpass filtering to avoid artifacts in the evaluated frequency range.
        Default is (None, None).
    hset_info : tuple[float, float, float], optional
        Tuple specifying the range of the resampling factors as (min, max, step). Default is (1.05, 2.0, 0.05).
    hset_accuracy : int, optional
        Decimal precision for the resampling factors. Default is 4.

    Returns
    -------
    IrasaSpectrum
        An object containing the following attributes:
            - freqs: np.ndarray
                Frequencies corresponding to the computed spectra.
            - raw_spectrum: np.ndarray
                The raw power spectrum.
            - aperiodic: np.ndarray
                The aperiodic (fractal) component of the spectrum.
            - periodic: np.ndarray
                The periodic (oscillatory) component of the spectrum.
            - ch_names: np.ndarray
                Channel names if provided.

    Notes
    -----
    This function provides fine-grained control over the IRASA parameters. For users working with MNE-Python,
    the `irasa_raw` and `irasa_epochs` functions from `pyrasa.irasa_mne` are recommended, as they handle
    additional preprocessing steps.

    References
    ----------
    Wen, H., & Liu, Z. (2016). Separating Fractal and Oscillatory Components in the Power Spectrum
    of Neurophysiological Signal. Brain Topography, 29(1), 13–26. https://doi.org/10.1007/s10548-015-0448-0
    """

    # set parameters
    if win_func_kwargs is None:
        win_func_kwargs = {}

    # Minimal safety checks
    if data.ndim == 1:
        data = data[np.newaxis, :]
    assert data.ndim == 2, 'Data shape needs to be either of shape (Channels, Samples) or (Samples, ).'  # noqa PLR2004

    irasa_params = {
        'data': data,
        'fs': fs,
        'band': band,
        'filter_settings': filter_settings,
        'hset_accuracy': hset_accuracy,
    }

    _check_irasa_settings(irasa_params=irasa_params, hset_info=hset_info)

    hset = np.round(np.arange(*hset_info), hset_accuracy)
    hset = [h for h in hset if h % 1 != 0]  # filter integers

    win_kwargs = {'win_func': win_func, 'win_func_kwargs': win_func_kwargs}
    dpss_settings = {
        'time_bandwidth': dpss_settings_time_bandwidth,
        'low_bias': dpss_settings_low_bias,
        'eigenvalue_weighting': dpss_eigenvalue_weighting,
    }

    def _local_irasa_fun(
        data: np.ndarray,
        fs: int,
        *args: Any,
        **kwargs: Any,
    ) -> np.ndarray:
        return _compute_psd_welch(
            data,
            fs=fs,
            nperseg=psd_kwargs.get('nperseg'),
            win_kwargs=win_kwargs,
            dpss_settings=dpss_settings,
            noverlap=psd_kwargs.get('noverlap'),
            nfft=psd_kwargs.get('nfft'),
        )[1]

    freq, psd = _compute_psd_welch(
        data,
        fs=fs,
        nperseg=psd_kwargs.get('nperseg'),
        win_kwargs=win_kwargs,
        dpss_settings=dpss_settings,
        noverlap=psd_kwargs.get('noverlap'),
        nfft=psd_kwargs.get('nfft'),
    )

    psd, psd_aperiodic, psd_periodic = _gen_irasa(
        data=np.squeeze(data), orig_spectrum=psd, fs=fs, irasa_fun=_local_irasa_fun, hset=hset
    )

    freq, psd_aperiodic, psd_periodic, psd = _crop_data(band, freq, psd_aperiodic, psd_periodic, psd, axis=-1)

    return IrasaSpectrum(
        freqs=freq,
        raw_spectrum=psd,
        aperiodic=psd_aperiodic,
        periodic=psd_periodic,
        ch_names=ch_names,
    )


# irasa sprint
def irasa_sprint(  # noqa PLR0915 C901
    data: np.ndarray,
    fs: int,
    ch_names: np.ndarray | None = None,
    band: tuple[float, float] = (1.0, 100.0),
    win_duration: float = 0.4,
    overlap_fraction: float = 0.90,
    win_func: Callable = dsp.windows.hann,
    win_func_kwargs: dict | None = None,
    dpss_settings_time_bandwidth: float = 2.0,
    dpss_settings_low_bias: bool = True,
    dpss_eigenvalue_weighting: bool = True,
    filter_settings: tuple[float | None, float | None] = (None, None),
    hset_info: tuple[float, float, float] = (1.05, 2.0, 0.05),
    hset_accuracy: int = 4,
) -> IrasaTfSpectrum:
    """
    Computes time-resolved aperiodic and periodic components of the power spectrum from a time series
    using the Irregular Resampling Autocorrelation (IRASA) algorithm.

    This function is useful for analyzing how the aperiodic and periodic components of the power spectrum
    change over time, providing a time-frequency decomposition of the signal.

    Parameters
    ----------
    data : np.ndarray
        Time series data, where the shape is expected to be either (Samples,) or (Channels, Samples).
    fs : int
        Sampling frequency of the data in Hz.
    ch_names : np.ndarray | None, optional
        Channel names associated with the data, if available. Default is None.
    band : tuple[float, float], optional
        The frequency range (lower and upper bounds in Hz) over which to compute the spectra. Default is (1.0, 100.0).
    win_duration : float, optional
        Duration of the window in seconds used for the short-time Fourier transforms (STFTs). Default is 0.4 seconds.
    overlap_fraction : int, optional
        The overlap between the STFT sliding windows as fraction. Default is .99 of the windows.
    win_func : Callable, optional
        Window function to be used in computing the time frequency spectrum. Default is `dsp.windows.hann`.
    win_func_kwargs : dict | None, optional
        Additional keyword arguments for the window function. Default is None.
    dpss_settings_time_bandwidth : float, optional
        Time-bandwidth product for the DPSS windows if used. Default is 2.0.
    dpss_settings_low_bias : bool, optional
        Keep only tapers with eigenvalues > 0.9. Default is True.
    dpss_eigenvalue_weighting : bool, optional
        Whether or not to apply eigenvalue weighting in DPSS. If True, spectral estimates weighted by
        the concentration ratio of their respective tapers before combining. Default is True.
    filter_settings : tuple[float | None, float | None], optional
        Cutoff frequencies for highpass and lowpass filtering to avoid artifacts in the evaluated frequency range.
        Default is (None, None).
    hset_info : tuple[float, float, float], optional
        Tuple specifying the range of the resampling factors as (min, max, step). Default is (1.05, 2.0, 0.05).
    hset_accuracy : int, optional
        Decimal precision for the resampling factors. Default is 4.

    Returns
    -------
    IrasaTfSpectrum
        An object containing the following attributes:
            - freqs: np.ndarray
                Frequencies corresponding to the computed spectra.
            - time: np.ndarray
                Time bins in seconds associated with the (a-)periodic spectra.
            - raw_spectrum: np.ndarray
                The raw time-frequency power spectrum.
            - aperiodic: np.ndarray
                The aperiodic (fractal) component of the spectrum.
            - periodic: np.ndarray
                The periodic (oscillatory) component of the spectrum.
            - ch_names: np.ndarray
                Channel names if provided.

    Notes
    -----
    This function performs a time-frequency decomposition of the input data, allowing for a time-resolved analysis
    of the periodic and aperiodic components of the signal. The STFT is computed for each time window, and IRASA
    is applied to separate the spectral components.

    References
    ----------
    Wen, H., & Liu, Z. (2016). Separating Fractal and Oscillatory Components in the Power Spectrum of
    Neurophysiological Signal. Brain Topography, 29(1), 13–26. https://doi.org/10.1007/s10548-015-0448-0
    """

    # set parameters
    if win_func_kwargs is None:
        win_func_kwargs = {}

    # Minimal safety checks
    if data.ndim == 1:
        data = data[np.newaxis, :]
    assert data.ndim == 2, 'Data shape needs to be either of shape (Channels, Samples) or (Samples, ).'  # noqa PLR2004

    irasa_params = {
        'data': data,
        'fs': fs,
        'band': band,
        'filter_settings': filter_settings,
    }

    _check_irasa_settings(irasa_params=irasa_params, hset_info=hset_info)

    hset = np.round(np.arange(*hset_info), hset_accuracy)
    hset = [h for h in hset if h % 1 != 0]  # filter integers

    win_kwargs = {'win_func': win_func, 'win_func_kwargs': win_func_kwargs}
    dpss_settings = {
        'time_bandwidth': dpss_settings_time_bandwidth,
        'low_bias': dpss_settings_low_bias,
        'eigenvalue_weighting': dpss_eigenvalue_weighting,
    }

    nfft = int(2 ** np.ceil(np.log2(np.max(hset) * win_duration * fs)))
    hop = int((1 - overlap_fraction) * win_duration * fs)
    # hop = int((1 - overlap_fraction) * nfft)
    irasa_kwargs: IrasaSprintKwargsTyped = {
        'nfft': nfft,
        'hop': hop,
        'win_duration': win_duration,
        'dpss_settings': dpss_settings,
        'win_kwargs': win_kwargs,
    }

    def _local_irasa_fun(
        data: np.ndarray,
        fs: int,
        h: float,
        up_down: str | None,
        time_orig: np.ndarray | None = None,
    ) -> np.ndarray:
        return _compute_sgramm(data, fs, h=h, up_down=up_down, time_orig=time_orig, **irasa_kwargs)[2]

    # get time and frequency info
    freq, time, sgramm = _compute_sgramm(data, fs, **irasa_kwargs)

    sgramm, sgramm_aperiodic, sgramm_periodic = _gen_irasa(
        data=data,
        orig_spectrum=sgramm,
        fs=fs,
        irasa_fun=_local_irasa_fun,
        hset=hset,
        time=time,
    )
    single_ch_dim = 2
    sgramm_aperiodic = (
        sgramm_aperiodic[np.newaxis, :, :] if sgramm_aperiodic.ndim == single_ch_dim else sgramm_aperiodic
    )
    sgramm_periodic = sgramm_periodic[np.newaxis, :, :] if sgramm_periodic.ndim == single_ch_dim else sgramm_periodic

    freq, sgramm_aperiodic, sgramm_periodic, sgramm = _crop_data(
        band, freq, sgramm_aperiodic, sgramm_periodic, sgramm, axis=1
    )

    # adjust time info (i.e. cut the padded stuff)
    tmax = data.shape[1] / fs
    t_mask = np.logical_and(time >= 0, time < tmax)
    freq_mask = freq > (1 / win_duration)  # mask rayleigh

    return IrasaTfSpectrum(
        freqs=freq[freq_mask],
        time=time[t_mask],
        raw_spectrum=sgramm,
        periodic=sgramm_periodic[:, :, t_mask][:, freq_mask, :],
        aperiodic=sgramm_aperiodic[:, :, t_mask][:, freq_mask, :],
        ch_names=ch_names,
    )



import numpy as np
import pytest

from pyrasa import irasa

from .settings import EXPONENT, FS, MIN_CORR_PSD_CMB, OSC_FREQ, TOLERANCE

# Estimate periodic and aperiodic components with IRASA
# These tests should cover the basic functionality of the IRASA workflow


# fixed slope
@pytest.mark.parametrize('exponent', EXPONENT, scope='session')
@pytest.mark.parametrize('osc_freq', OSC_FREQ, scope='session')
@pytest.mark.parametrize('fs', FS, scope='session')
def test_irasa(combined_signal, fs, osc_freq, exponent):
    f_range = [1, 100]
    irasa_spectrum = irasa(combined_signal, fs, f_range, psd_kwargs={'nperseg': 4 * fs})
    # test the shape of the output
    assert irasa_spectrum.freqs.shape[0] == irasa_spectrum.aperiodic.shape[1] == irasa_spectrum.periodic.shape[1]
    # test the selected frequency range
    assert bool(np.logical_and(irasa_spectrum.freqs[0] == f_range[0], irasa_spectrum.freqs[-1] == f_range[1]))
    # test whether recombining periodic and aperiodic spectrum is equivalent to the original spectrum
    # freqs_psd, psd = dsp.welch(combined_signal, fs, nperseg=int(4 * fs))
    psd_cmb = irasa_spectrum.aperiodic[0, :] + irasa_spectrum.periodic[0, :]
    # freq_logical = np.logical_and(freqs_psd >= f_range[0], freqs_psd <= f_range[1])
    r = np.corrcoef(irasa_spectrum.raw_spectrum, psd_cmb)[0, 1]
    assert r > MIN_CORR_PSD_CMB
    # test whether we can reconstruct the exponent correctly
    aperiodic_fit = irasa_spectrum.fit_aperiodic_model(fit_func='fixed')
    assert bool(np.isclose(aperiodic_fit.aperiodic_params['Exponent'][0], np.abs(exponent), atol=TOLERANCE))
    # test whether we can reconstruct the peak frequency correctly
    pe_params = irasa_spectrum.get_peaks()
    assert bool(np.isclose(np.round(pe_params['cf'], 0), osc_freq))
